# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for MetricManager and ShardedEvaluator in seqio.evaluation."""

from typing import Any, Mapping, Optional, Sequence
from unittest import mock

from absl.testing import absltest
import flax
import numpy as np
from seqio import dataset_providers
from seqio import evaluation
from seqio import metrics as metrics_lib
from seqio import utils
from seqio import vocabularies

ModelOutputType = metrics_lib.ModelOutputType


@flax.struct.dataclass
class Count(metrics_lib.Metric):
  """Implements a simple counting metric, which inherits seqio.Metric."""

  count: int = 0
  model_output_type: ModelOutputType = ModelOutputType.PREDICTION

  @classmethod
  def empty(cls) -> "Count":
    return cls(count=0)

  @classmethod
  def from_model_output(
      cls,
      inputs: Sequence[Mapping[str, Any]],
      model_output: np.ndarray,
      features: Mapping[str, utils.Feature],
      target_field_name: str = "targets",
      mask: Optional[np.ndarray] = None,
      indices_2d: Optional[np.ndarray] = None,
  ) -> "Count":

    if mask is None:
      mask = np.ones((len(inputs),))

    return cls(count=mask.sum())

  def merge(self, other: "Count") -> "Count":
    """Returns `Count` that is the accumulation of `self` and `other`.

    Args:
      other: A `Count` whose inermediate values should be accumulated onto the
        values of `self`. Note that in a distributed setting, `other` will
        typically be the output of a `jax.lax` parallel operator and thus have a
        dimension added to the dataclass returned by `.from_model_output()`.

    Returns:
      A new `Count` instance that accumulates the value from both `self` and
        `other`.
    """
    return type(self)(count=self.count + other.count)

  def compute(self):
    return {"count": self.count}


class MetricManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.uncalled_fn = mock.Mock()

    self.task = mock.Mock()
    self.task.name = "dummy_task"
    mf0 = lambda targets, predictions: {"accuracy": 1.0}
    mf1 = lambda targets, predictions: {"loss": 0.3}
    self.task.metric_objs = [
        metrics_lib.PassthroughLegacyMetric.from_metric_fn(mf0, None).empty(),
        metrics_lib.PassthroughLegacyMetric.from_metric_fn(mf1, None).empty(),
        Count.empty(),
    ]

    mock_vocab = mock.create_autospec(vocabularies.Vocabulary)
    self.task.output_features = {
        "targets": dataset_providers.Feature(mock_vocab)
    }

    self.metric_manager = evaluation.MetricManager([self.task])

  def tearDown(self):
    super().tearDown()
    self.uncalled_fn.assert_not_called()

  def test_initialize_metrics(self):
    model_output_type = metrics_lib.ModelOutputType.PREDICTION
    self.metric_manager.initialize_metrics(self.task.name, model_output_type)

    self.assertSequenceEqual(
        list(self.metric_manager.output_metrics_collections.keys()),
        ["dummy_task"],
    )
    self.assertSequenceEqual(
        list(
            self.metric_manager.output_metrics_collections["dummy_task"].keys()
        ),
        [model_output_type],
    )

    metrics = self.metric_manager.output_metrics_collections["dummy_task"][
        model_output_type
    ]
    self.assertEqual(metrics.FromMetricFun_0.values, {})
    self.assertEqual(metrics.FromMetricFun_1.values, {})
    self.assertEqual(metrics.Count_2.count, 0)

  def test_from_model_output(self):
    task_batch = [
        {"targets": [0, 1, 2, 3]},
        {"targets": [1, 2, 3, 4]},
    ]

    batch_result = np.array([[0, 1, 2, 3], [1, 2, 3, 4]])

    batch_indices_2d = np.array([
        [0, 0],  # 1st example is from shard 0 and index 0
        [0, 1],  # 1st example is from shard 0 and index 1
    ])
    model_output_type = metrics_lib.ModelOutputType.PREDICTION
    metrics_batch = self.metric_manager.from_model_output(
        inputs=task_batch,
        model_output=batch_result,
        features=self.task.output_features,
        mask=(batch_indices_2d[:, 1] >= 0),
        indices_2d=batch_indices_2d,
        task_name=self.task.name,
        model_output_type=model_output_type,
    )

    self.assertEqual(metrics_batch.Count_2.count, 2)

  def test_merge(self):
    model_output_type = metrics_lib.ModelOutputType.PREDICTION
    self.metric_manager.initialize_metrics(self.task.name, model_output_type)

    task_batch = [
        {"targets": [0, 1, 2, 3]},
        {"targets": [1, 2, 3, 4]},
    ]
    batch_result = np.array([[0, 1, 2, 3], [1, 2, 3, 4]])
    batch_indices_2d = np.array([
        [0, 0],  # 1st example is from shard 0 and index 0
        [0, 1],  # 1st example is from shard 0 and index 1
    ])
    metrics_batch = self.metric_manager.from_model_output(
        inputs=task_batch,
        model_output=batch_result,
        features=self.task.output_features,
        mask=(batch_indices_2d[:, 1] >= 0),
        indices_2d=batch_indices_2d,
        task_name=self.task.name,
        model_output_type=model_output_type,
    )

    self.metric_manager.merge(
        metrics_batch,
        task_name=self.task.name,
        model_output_type=model_output_type,
    )
    metrics = self.metric_manager.output_metrics_collections["dummy_task"][
        model_output_type
    ]
    # The first two metrics are CollectingMetric's, so they are just passing
    # through model outputs.
    np.testing.assert_array_equal(
        metrics.FromMetricFun_0.values["indices_2d"], batch_indices_2d
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_0.values["mask"], np.array([True, True])
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_0.values["model_output"], batch_result
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_1.values["indices_2d"], batch_indices_2d
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_1.values["mask"], np.array([True, True])
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_1.values["model_output"], batch_result
    )
    # The last metric is batch-update enabled metric.
    self.assertEqual(metrics.Count_2.count, 2)

    # Merge a second time, so that everything in the metric objects should be
    # doubled in terms of the number of items for CollectingMetric or scalar
    # size for Count metric.
    self.metric_manager.merge(
        metrics_batch,
        task_name=self.task.name,
        model_output_type=model_output_type,
    )
    metrics = self.metric_manager.output_metrics_collections["dummy_task"][
        model_output_type
    ]

    # The first two metrics are CollectingMetric's, so they are just passing
    # through model outputs.
    np.testing.assert_array_equal(
        metrics.FromMetricFun_0.values["indices_2d"],
        np.concatenate([batch_indices_2d, batch_indices_2d], axis=0),
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_0.values["mask"],
        np.array([True, True, True, True]),
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_0.values["model_output"],
        np.concatenate([batch_result, batch_result], axis=0),
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_1.values["indices_2d"],
        np.concatenate([batch_indices_2d, batch_indices_2d], axis=0),
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_1.values["mask"],
        np.array([True, True, True, True]),
    )
    np.testing.assert_array_equal(
        metrics.FromMetricFun_1.values["model_output"],
        np.concatenate([batch_result, batch_result], axis=0),
    )
    # The last metric is batch-update enabled metric.
    self.assertEqual(metrics.Count_2.count, 4)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MetricValue objects to wrap results being returned by metric funcitons."""

import dataclasses
import enum
import inspect
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import clu.metrics
import flax
import jax
import jax.numpy as jnp
import numpy as np
from seqio import utils
import tensorflow.compat.v2 as tf


@dataclasses.dataclass
class MetricValue:
  """A base method for the dataclasses that represent tensorboard values.

  Task `metric_fn`s should output `Mapping[str, MetricValue]` which will be
  written by a `Logger`.
  """


@dataclasses.dataclass
class Scalar(MetricValue):
  """The default tensorflow value, used for creating time series graphs."""

  value: Union[int, float]


@dataclasses.dataclass
class Text(MetricValue):
  """Text to output to tensorboard, markdown is rendered by tensorboard."""

  textdata: Union[str, bytes]


@dataclasses.dataclass
class Image(MetricValue):
  """An image to output to tensorboard.

  The format for the image array should match the format expected for the data
  parameter described
  [here](https://www.tensorflow.org/api_docs/python/tf/summary/image).
  """

  image: np.ndarray
  max_outputs: int = 3


@dataclasses.dataclass
class Audio(MetricValue):
  """An audio example to output to tensorboard.

  The format for the audio array should match the format expected for the data
  parameter described
  [here](https://www.tensorflow.org/api_docs/python/tf/summary/audio).
  """

  audiodata: np.ndarray
  sample_rate: int = 44100
  max_outputs: int = 3


@dataclasses.dataclass
class Histogram(MetricValue):
  """A histogram to output to tensorboard."""

  values: np.ndarray
  bins: Optional[int] = None


@dataclasses.dataclass
class Generic(MetricValue):
  """A raw tensor to output to tensorboard."""

  tensor: np.ndarray
  metadata: tf.compat.v1.SummaryMetadata


class ModelOutputType(enum.IntEnum):
  """Model output types."""

  PREDICTION = 1
  SCORE = 2
  PREDICTION_WITH_AUX = 3
  SCORE_WITH_INTERMEDIATES = 4

  @classmethod
  def to_str(cls, enm):
    return {
        cls.PREDICTION: "prediction",
        cls.SCORE: "score",
        cls.PREDICTION_WITH_AUX: "prediction_with_aux",
        cls.SCORE_WITH_INTERMEDIATES: "score_with_intermediates",
    }[enm]


MetricFnCallable = Callable[..., Mapping[str, Union[MetricValue, float]]]


@flax.struct.dataclass
class Metric(clu.metrics.Metric):
  """Base Metric class for seqio evaluation."""

  model_output_type: ModelOutputType

  @classmethod
  def from_model_output(
      cls,
      inputs: Sequence[Mapping[str, Any]],
      model_output: Union[np.ndarray, Tuple[np.ndarray, np.ndarray]],
      features: Mapping[str, utils.Feature],
      target_field_name: str = "targets",
      mask: Optional[np.ndarray] = None,
      indices_2d: Optional[np.ndarray] = None,
  ) -> "Metric":
    """Creates a `seqio.Metric` from model outputs.

    Args:
      inputs: Examples in dataset.
      model_output: Model output computed by model functions.
      features: Output features defined in seqio.Task.
      target_field_name: Field name of the target sequence.
      mask: A boolean array to indicate which examples in the inputs are
        included for metric evaluation.
      indices_2d: 2d-indices of examples in the inputs/model_output. First
        dimension is shard id, the second is the example id within that shard.

    Returns:
      An instance of Metric.
    Raises:
      NotImplementedError: Must override from_model_output()
    """

    raise NotImplementedError("Must override from_model_output()")


class CollectingMetric(clu.metrics.CollectingMetric):
  """CollectingMetric interface for seqio evaluation."""

  @classmethod
  def from_model_output(  # pylint:disable=missing-function-docstring
      cls,
      inputs: Sequence[Mapping[str, Any]],
      model_output: Union[np.ndarray, Tuple[np.ndarray, np.ndarray]],
      features: Mapping[str, utils.Feature],
      target_field_name: str = "targets",
      mask: Optional[np.ndarray] = None,
      indices_2d: Optional[np.ndarray] = None,
  ):
    del inputs, features, target_field_name
    num_examples = (
        len(model_output[0])
        if isinstance(model_output, tuple)
        else len(model_output)
    )

    if mask is None:
      mask = jnp.ones((num_examples,), jnp.int32)

    if indices_2d is None:
      indices_2d = jnp.transpose(
          jnp.stack([
              jnp.zeros((num_examples,), jnp.int32),
              jnp.arange(num_examples, dtype=jnp.int32),
          ])
      )
    return cls(
        values={
            "model_output": model_output,
            "indices_2d": indices_2d,
            "mask": mask,
        }
    )

  def actual_compute(
      self,
      task_dataset_as_numpy,
      task_output_features,
      target_field_name: str = "targets",
      cached_targets: Optional[List[str]] = None,
  ):
    """Implements the metric computation logics for CollectingMetric.

    Args:
      task_dataset_as_numpy: Examples in dataset.
      task_output_features: Output features defined in the seqio.Task.
      target_field_name: Field name of the target sequence.
      cached_targets: targets that have been cached by Evaluator and can be
        supplied here to save time of post-processing targets.

    Returns:
      A tuple of two items, first item is a dict of metric results, the second
        item is targets_and_inferences.
    Raises:
      NotImplementedError: Must override from_model_output()
    """

    raise NotImplementedError("Must override from_model_output()")


# TODO(kehanghan): consider using CollectingMetric for LegacyMetric.
@flax.struct.dataclass
class LegacyMetric(Metric):
  """Metric class for legacy use-case where metric fn is supplied."""

  _metric_fn: MetricFnCallable
  _postprocess_fn: Callable[..., Any]
  metric_fn_kwargs: Dict[str, Any]
  targets_and_inferences: Dict[str, Any]

  @classmethod
  def empty(cls, metric_fn, postprocess_fn) -> "LegacyMetric":
    pos_args = tuple(
        key
        for key, param in inspect.signature(metric_fn).parameters.items()
        if param.default == inspect.Parameter.empty
        and param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
    )
    if pos_args == ("targets", "scores"):
      model_output_type = ModelOutputType.SCORE
    elif pos_args == ("targets", "predictions"):
      model_output_type = ModelOutputType.PREDICTION
    elif pos_args == ("targets", "predictions", "aux_values"):
      model_output_type = ModelOutputType.PREDICTION_WITH_AUX
    else:
      raise ValueError(
          "Metric functions must have positional arguments matching either "
          "('targets', 'scores'), ('targets', 'predictions') or "
          "('targets', 'predictions', 'aux_values'). "
          f"Got: {pos_args}"
      )

    return cls(
        _metric_fn=metric_fn,
        _postprocess_fn=postprocess_fn,
        model_output_type=model_output_type,
        metric_fn_kwargs={},
        targets_and_inferences={},
    )

  def postprocess_fn(
      self, targets_or_predictions: Any, **postprocess_kwargs
  ) -> Any:
    """Applies the postprocessing to targets or predictions."""
    if self._postprocess_fn:
      return self._postprocess_fn(targets_or_predictions, **postprocess_kwargs)
    return targets_or_predictions

  def from_model_output(  # pylint:disable=arguments-renamed
      self,
      inputs: Sequence[Mapping[str, Any]],
      model_output: Union[np.ndarray, Tuple[np.ndarray, np.ndarray]],
      features: Mapping[str, utils.Feature],
      target_field_name: str = "targets",
      mask: Optional[np.ndarray] = None,
  ) -> "LegacyMetric":
    if not self.metric_fn_kwargs.get("targets"):
      # Postprocesses the targets here.
      postprocessed_targets = []
      for ex in inputs:
        pretokenized_target_field_name = target_field_name + "_pretokenized"
        if pretokenized_target_field_name in ex:
          target = ex[pretokenized_target_field_name]
        else:
          target = features[target_field_name].vocabulary.decode(
              list(ex[target_field_name])
          )
        if isinstance(target, bytes):
          target = target.decode("utf-8")

        postprocessed_targets.append(
            self.postprocess_fn(target, example=ex, is_target=True)
        )
      self.metric_fn_kwargs["targets"] = postprocessed_targets
      self.targets_and_inferences["targets"] = postprocessed_targets

    if self.model_output_type == ModelOutputType.SCORE:
      self.metric_fn_kwargs["scores"] = model_output
      self.targets_and_inferences["score"] = model_output
    else:
      vocab = features[target_field_name].vocabulary
      if self.model_output_type == ModelOutputType.PREDICTION_WITH_AUX:
        self.metric_fn_kwargs["aux_values"] = model_output[1]
        self.targets_and_inferences["aux_value"] = model_output[1]
        predictions = [vocab.decode(tokens) for tokens in model_output[0]]
      elif self.model_output_type == ModelOutputType.PREDICTION:
        # Default behavior for top-1 decoding, model_output is a 2d array.
        # first dim is for batch, second is for sequence length.
        if isinstance(model_output, np.ndarray) and model_output.ndim == 2:
          predictions = [vocab.decode(tokens) for tokens in model_output]
        else:
          # In case of top-k decoding, model_output will be a 3d array
          # first dim is for batch, second is for num_decodes, third is for
          # sequence length.
          predictions = []
          for sequences in model_output:
            predictions_for_one_example = []
            for sequence in sequences:
              predictions_for_one_example.append(vocab.decode(sequence))
            predictions.append(predictions_for_one_example)
        self.targets_and_inferences["output"] = predictions

      # Postprocesses the predictions here.
      postprocessed_predictions = [
          self.postprocess_fn(p, example=ex, is_target=False)
          for ex, p in zip(inputs, predictions)
      ]

      self.metric_fn_kwargs["predictions"] = postprocessed_predictions
      self.targets_and_inferences["prediction"] = postprocessed_predictions

    return self

  def compute(self):
    return self._metric_fn(**self.metric_fn_kwargs)


def remove_padding_examples(model_output, indices_2d, mask):
  """Removes padding examples indicated by the mask array.

  Args:
    model_output: model outputs of all the examples (including the padding
      ones). The padding examples are used to make sure during inference, the
      inference function receives full batch if the last batch does not enough
      examples.
    indices_2d: 2d indices of all the examples.
    mask: an array of booleans. 1 indicates valid example, 0 indicates padded
      example that needs to be removed.

  Returns:
    2d-indices and model outputs of all the non-padding examples.
  """
  indices_2d = indices_2d[mask == 1]
  model_output = jax.tree.map(lambda x: x[mask == 1], model_output)

  return indices_2d, model_output


def globally_sort_model_output(model_output, indices_2d):
  """Globally sorts model ouputs by the 2d indices of the examples.

  The sorting is done first by shard id (first index of the 2d-index) and then
  by example id (second index of the 2d-index).

  Args:
    model_output: model outputs of all the examples.
    indices_2d: 2d indices of all the examples.

  Returns:
    sorted model outputs.
  """
  permutation = np.lexsort((indices_2d[:, 1], indices_2d[:, 0]))

  def _sort_by_permutation(x):
    return np.array([x[permutation[i]] for i in range(len(permutation))])

  model_output = jax.tree.map(_sort_by_permutation, model_output)

  return model_output


@flax.struct.dataclass
class PassthroughLegacyMetric(CollectingMetric):
  """Makes PassthroughLegacyMetric from metric functions."""

  @classmethod
  def from_metric_fn(
      cls,
      metric_fn: MetricFnCallable,
      postprocess_fn: Optional[Callable[..., Any]] = None,
  ):
    """Creates `PassthroughLegacyMetric` from `metric_fn` and `postprocess_fn`.

    Example:

    ```
    squad_cls = PassthroughLegacyMetric.from_metric_fn(
        metric_fn=t5_metrics.squd, postprocess_fn=t5_postprocessors.qa)
    ```

    Args:
      metric_fn: Function used to compute metric.
      postprocess_fn: Function used to process targets (vocab decoded) and
        predictions (vocab decoded) before feeding into metric_fn.

    Returns:
      A `Metric` that calls `metric_fn` and `postprocess_fn` in its
      `.from_model_output()`.
    """

    def _get_model_output_type() -> ModelOutputType:
      pos_args = tuple(
          key
          for key, param in inspect.signature(metric_fn).parameters.items()
          if param.default == inspect.Parameter.empty
          and param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
      )
      if pos_args == ("targets", "scores"):
        model_output_type = ModelOutputType.SCORE
      elif pos_args == ("targets", "predictions"):
        model_output_type = ModelOutputType.PREDICTION
      elif pos_args == ("targets", "predictions", "aux_values"):
        model_output_type = ModelOutputType.PREDICTION_WITH_AUX
      else:
        raise ValueError(
            "Metric functions must have positional arguments matching either "
            "('targets', 'scores'), ('targets', 'predictions') or "
            "('targets', 'predictions', 'aux_values'). "
            f"Got: {pos_args}"
        )

      return model_output_type

    @flax.struct.dataclass
    class FromMetricFun(cls):
      """Wrapper PassthroughLegacyMetric class that runs metric_fn."""

      model_output_type: ModelOutputType = _get_model_output_type()

      @classmethod
      def postprocess(
          cls, targets_or_predictions: Any, **postprocess_kwargs
      ) -> Any:
        """Applies the postprocessing to targets or predictions."""
        if postprocess_fn:
          return postprocess_fn(targets_or_predictions, **postprocess_kwargs)
        return targets_or_predictions

      def postprocess_targets(
          self,
          task_dataset_as_numpy,
          task_output_features,
          target_field_name: str = "targets",
      ):
        """Applies the postprocessing to targets."""
        # Postprocesses the targets here.
        postprocessed_targets = []
        for ex in task_dataset_as_numpy:
          pretokenized_target_field_name = target_field_name + "_pretokenized"
          if pretokenized_target_field_name in ex:
            target = ex[pretokenized_target_field_name]
          else:
            target = task_output_features[target_field_name].vocabulary.decode(
                list(ex[target_field_name])
            )
          if isinstance(target, bytes):
            target = target.decode("utf-8")

          postprocessed_targets.append(
              type(self).postprocess(target, example=ex, is_target=True)
          )
        return postprocessed_targets

      def actual_compute(
          self,
          task_dataset_as_numpy,
          task_output_features,
          target_field_name: str = "targets",
          cached_targets: Optional[List[str]] = None,
      ):
        # Postprocesses the targets here.
        if not cached_targets:
          postprocessed_targets = self.postprocess_targets(
              task_dataset_as_numpy, task_output_features, target_field_name
          )
        else:
          postprocessed_targets = cached_targets

        metric_fn_kwargs, targets_and_inferences = {}, {}
        metric_fn_kwargs["targets"] = postprocessed_targets
        targets_and_inferences["targets"] = postprocessed_targets

        # We process the model outputs here by the steps below.
        # Step 1: removes padded examples using mask.
        indices_2d, model_output = remove_padding_examples(
            self.values["model_output"],
            self.values["indices_2d"],
            self.values["mask"],
        )

        assert len(postprocessed_targets) == len(indices_2d)

        # Step 2: sorts the model outputs by 2d-indices, namely (shard_id,
        # index_within_shard) to align with targets.
        model_output = globally_sort_model_output(model_output, indices_2d)

        if type(self).model_output_type == ModelOutputType.SCORE:
          metric_fn_kwargs["scores"] = model_output
          targets_and_inferences["score"] = model_output
        else:
          vocab = task_output_features[target_field_name].vocabulary
          if (
              type(self).model_output_type
              == ModelOutputType.PREDICTION_WITH_AUX
          ):
            metric_fn_kwargs["aux_values"] = model_output[1]
            targets_and_inferences["aux_value"] = model_output[1]
            predictions = [vocab.decode(tokens) for tokens in model_output[0]]
          elif type(self).model_output_type == ModelOutputType.PREDICTION:
            # Default behavior for top-1 decoding, model_output is a 2d array.
            # first dim is for batch, second is for sequence length.
            if isinstance(model_output, np.ndarray) and model_output.ndim == 2:
              predictions = [vocab.decode(tokens) for tokens in model_output]
            elif (
                isinstance(model_output, np.ndarray) and model_output.ndim == 3
            ):
              # In case of top-k decoding, model_output will be a 3d array
              # first dim is for batch, second is for num_decodes, third is for
              # sequence length.
              predictions = []
              for sequences in model_output:
                predictions_for_one_example = []
                for sequence in sequences:
                  predictions_for_one_example.append(vocab.decode(sequence))
                predictions.append(predictions_for_one_example)
            else:
              # If neither 2d or 3d, assume that model_output is already
              # decoded.
              predictions = model_output
          targets_and_inferences["output"] = predictions

          # Postprocesses the predictions here.
          postprocessed_predictions = [
              type(self).postprocess(p, example=ex, is_target=False)
              for ex, p in zip(task_dataset_as_numpy, predictions)
          ]

          metric_fn_kwargs["predictions"] = postprocessed_predictions
          targets_and_inferences["prediction"] = postprocessed_predictions

        return metric_fn(**metric_fn_kwargs), targets_and_inferences

    return FromMetricFun



# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for the class-based evaluation."""

import collections
import concurrent
import functools
import inspect
import itertools
import time
from typing import Any, Callable, Mapping, MutableMapping, Optional, Sequence, Tuple, Type, Union

from absl import logging
import clu.metrics
import jax
from jax.experimental import multihost_utils
import numpy as np
from seqio import dataset_providers
from seqio import feature_converters
from seqio import loggers as loggers_lib
from seqio import metrics as metrics_lib
from seqio import utils
import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds
import typing_extensions

Task = dataset_providers.Task
ModelOutputType = metrics_lib.ModelOutputType
EncDecFeatureConverter = feature_converters.EncDecFeatureConverter
FeatureConverter = feature_converters.FeatureConverter

AllOutputTokensType = Mapping[str, Sequence[Sequence[int]]]
AllOutputScoresType = Any  # Mapping[str, Sequence[float]]
AllOutputAuxValuesType = Mapping[str, Mapping[str, Sequence[Any]]]
TaskMetricsType = MutableMapping[str, Any]
AllMetricsType = MutableMapping[str, TaskMetricsType]


class AllMetricsFuture(typing_extensions.Protocol):

  def result(self) -> AllMetricsType:
    ...


MetricsAndOutputsType = Tuple[AllMetricsFuture, Any]  # metrics  # outputs


def get_valid_eval_tasks(tasks: Sequence[Task], split: str) -> Sequence[Task]:
  """Get tasks that have the specified split and a metric function."""

  valid_tasks = []

  for task in tasks:
    if split not in task.splits:
      logging.info(
          "Task %s has no '%s' split; skipping eval.", task.name, split
      )
      continue
    if not task.metric_fns and not task.metric_objs:
      logging.info("Task %s has no metrics defined; skipping eval.", task.name)
      continue
    metric_types = []
    if task.predict_metric_fns:
      metric_types.append("predict")
    if task.predict_with_aux_metric_fns:
      metric_types.append("predict_with_aux")
    if task.score_metric_fns:
      metric_types.append("score")
    logging.info(
        "Adding task '%s' with %s metric_fn(s).",
        task.name,
        " and ".join(metric_types),
    )
    valid_tasks.append(task)

  return valid_tasks


def _cache_and_measure_examples(
    tasks: Sequence[Task],
    dataset_fn: Callable[[Task], tf.data.Dataset],
    sequence_dims: Mapping[str, int],
    num_examples: Optional[int] = None,
    use_memory_cache: bool = True,
) -> Tuple[Mapping[str, tf.data.Dataset], Mapping[str, int]]:
  """Get cached datasets and maximum sequence lengths per feature.

  Args:
    tasks: tasks objects to get targets and examples for.
    dataset_fn: function, returns the dataset from the task object.
    sequence_dims: dict of feature names to their sequence dimension.
    num_examples: an optional maximum number of examples to take from the
      beginning of each task dataset.
    use_memory_cache: whether to use tf.data.Dataset#cache. may cause memory
      issues for large datasets.

  Returns:
    cached_task_datasets: cached datasets for each task, with cardinality set
    max_sequence_length: maximum sequence lengths for inputs and targets across
      all tasks.
  """
  # Pre-load in all of the targets once before entering continuous eval loop
  cached_task_datasets = {}
  max_sequence_length = {k: 0 for k in tasks[0].output_features.keys()}

  for task in tasks:
    assert (
        max_sequence_length.keys() == task.output_features.keys()
    ), "all tasks must have the same features"

  for task in tasks:
    ds = dataset_fn(task)
    if num_examples:
      ds = ds.take(num_examples)
    if use_memory_cache:
      ds = ds.cache()

    cnt = 0
    for ex in tfds.as_numpy(ds):
      for k in max_sequence_length:
        sequence_dim = sequence_dims.get(k, 0)
        if isinstance(ex[k], tf.RaggedTensor):
          sequence_length = tf.reduce_max(
              ex[k].row_lengths(axis=sequence_dim)
          ).numpy()
        else:
          sequence_length = ex[k].shape[sequence_dim]
        max_sequence_length[k] = max(max_sequence_length[k], sequence_length)
      cnt += 1

    cached_task_datasets[task.name] = ds.apply(
        tf.data.experimental.assert_cardinality(cnt)
    )

  return cached_task_datasets, max_sequence_length


_BatchId = int
_Tokens = Sequence[int]
_AuxValues = Mapping[str, Sequence[Any]]

_IndicesAndPredictions = Sequence[Tuple[_BatchId, _Tokens]]
_IndicesAndPredictionsWithAuxValues = Tuple[_IndicesAndPredictions, _AuxValues]
PredictFnReturnType = Union[
    _IndicesAndPredictions, _IndicesAndPredictionsWithAuxValues
]

_IndicesAndScores = Sequence[Tuple[_BatchId, float]]
_IndicesAndScoresWithIntermediates = Tuple[_IndicesAndScores, Mapping[str, Any]]
ScoreFnReturnType = Union[_IndicesAndScores, _IndicesAndScoresWithIntermediates]
ModelFnReturnType = Union[
    _IndicesAndPredictions,
    _IndicesAndPredictionsWithAuxValues,
    _IndicesAndScores,
    _IndicesAndScoresWithIntermediates,
]


class PredictFnCallable(typing_extensions.Protocol):

  def __call__(
      self,
      dataset: tf.data.Dataset,
      model_feature_shapes: Optional[Mapping[str, int]],
  ) -> PredictFnReturnType:
    ...


class ScoreFnCallable(typing_extensions.Protocol):

  def __call__(
      self,
      dataset: tf.data.Dataset,
      model_feature_shapes: Optional[Mapping[str, int]],
  ) -> ScoreFnReturnType:
    ...


class ModelFnCallable(typing_extensions.Protocol):

  def __call__(
      self,
      dataset: tf.data.Dataset,
      model_feature_shapes: Optional[Mapping[str, int]],
  ) -> ModelFnReturnType:
    ...


def _extract_model_output(cached_model_dataset, model_fn):
  """Extracts model output from a cached dataset."""

  def _permute(x, sorted_order):
    return [x[sorted_order[i]] for i in range(len(sorted_order))]

  model_fn_result = model_fn(cached_model_dataset)  # pytype: disable=missing-parameter  # always-use-return-annotations
  if isinstance(model_fn_result, tuple):
    # Some of model functions return a tuple of two outputs per example.
    # e.g., ModelOutputType.PREDICTION_WITH_AUX,
    # ModelOutputType.SCORE_WITH_INTERMEDIATES
    indices_and_outputs, all_aux_values = model_fn_result
    indices, outputs = zip(*indices_and_outputs)
    sorted_order = np.argsort(indices)

    sorted_outputs = _permute(outputs, sorted_order)
    sorted_aux = jax.tree.map(
        functools.partial(_permute, sorted_order=sorted_order),
        all_aux_values,
        is_leaf=lambda x: isinstance(x, list),
    )
    return sorted_outputs, sorted_aux

  else:
    # Majority of model functions should only return one output per example.
    # e.g., ModelOutputType.SCORE, ModelOutputType.PREDICTION
    indices_and_outputs = model_fn_result
    _, sorted_outputs = zip(*sorted(indices_and_outputs, key=lambda x: x[0]))
    return list(sorted_outputs)


def _extract_tokens_and_aux_values(cached_model_dataset, predict_fn):
  """Extracts tokens and aux scores from a cached dataset."""
  predict_fn_result = predict_fn(cached_model_dataset)

  all_aux_values = {}
  if isinstance(predict_fn_result, tuple):
    indices_and_tokens, all_aux_values = predict_fn_result
    indices, tokens = zip(*indices_and_tokens)

    permutation = np.argsort(indices)

    tokens = [tokens[permutation[i]] for i in range(len(permutation))]
    for aux_keys, aux_values in all_aux_values.items():
      all_aux_values[aux_keys] = [
          aux_values[permutation[i]] for i in range(len(permutation))
      ]

  else:
    indices_and_tokens = predict_fn_result
    _, tokens = zip(*sorted(indices_and_tokens, key=lambda x: x[0]))

  return tokens, all_aux_values


def _extract_scores(cached_model_dataset, score_fn):
  """Extracts scores and intermediate values from a cached dataset."""
  indices_and_scores = score_fn(cached_model_dataset)

  if isinstance(indices_and_scores, tuple) and len(indices_and_scores) == 2:
    indices_and_scores, intermediates = indices_and_scores
    indices, scores = zip(*indices_and_scores)
    sorted_order = np.argsort(indices)

    def _permute(x):
      return [x[sorted_order[i]] for i in range(len(sorted_order))]

    sorted_scores = _permute(scores)
    sorted_intermediates = jax.tree.map(
        _permute, intermediates, is_leaf=lambda x: isinstance(x, list)
    )
    return sorted_scores, sorted_intermediates

  if len(indices_and_scores[0]) != 2:
    raise ValueError(
        "Expected a sequence of length-2 tuples with (index, score) format."
    )
  return [x[1] for x in sorted(indices_and_scores, key=lambda x: x[0])]


class Evaluator:
  """A class to encapsulate all eval-related information.

  Users should define `predict_fn` and then pass it to `evaluate` method.
  `predict_fn` should operate with enumerated tf.data.Dataset. See `evaluate`
  method for more detail.

  evaluation data is cached once and will be used for arbitrary number of
  evaluation runs.

  If none of the evaluation tasks has metrics functions defined, the evaluation
  will be skipped. `Evaluator.evaluate` will return ({}, {}) assuming that
  compute_metrics is True.

  Note that we cache two versions of the datasets. The first version
  (self.cached_task_datasets) has the task features (e.g., "inputs" and
  "targets"), which are returned from `seqio.Task.get_dataset`. The second
  version (self.cached_model_datasets) has model features (e.g.,
  "decoder_target_tokens"). This is returned from the feature converter. The
  former is used for postprocessing associated with the Task that requires the
  original task datasets. The latter is passed to `predict_fn` for evaluation.

  Attributes:
    eval_tasks: a mapping from a mixture or a task name to seqio.Task object(s).
    cached_model_datasets: cached evaluation datasets with model features.
    cached_task_datasets: cached evaluation datasets with task features.
    model_feature_shapes: mapping from model feature to its shape in the
      `cached_model_datasets`.
    loggers: a sequence of subclasses of `Logger`.
  """

  def __init__(
      self,
      mixture_or_task_name: str,
      feature_converter: FeatureConverter,
      eval_split: str = "validation",
      use_cached: bool = False,
      seed: Optional[int] = 42,
      sequence_length: Optional[Mapping[str, int]] = None,
      num_examples: Optional[int] = None,
      shuffle: bool = False,
      logger_cls: Sequence[Type[loggers_lib.Logger]] = (),
      log_dir: Optional[str] = None,
      use_memory_cache: bool = True,
      async_compute_metrics: bool = True,
      target_field_name: str = "targets",
  ):
    """Evaluator constructor.

    Args:
      mixture_or_task_name: A registered task or mixture name.
      feature_converter: A feature converter object to use to convert the task
        features to model features. Must be a subclass of
        seqio.FeatureConverter.
      eval_split: The evaluation split name. Typically "validation" or "test".
      use_cached: Whether to use the cached dataset instead of processing it on
        the fly.
      seed: Random seed used for dataset shuffle and preprocessing. This is
        usually not needed since eval datasets aren't shuffled and shouldn't use
        stochastic operations. It is only useful for in certain data sources
        such as `FewshotDataSource` where the training examples are randomly
        selected during evaluation.
      sequence_length: An optional length specification. If specified, these
        will be the hard-limit on the evaluation data used for prediction. If
        none of the preprocessors depend on the sequence length, it can be left
        unspecified and the maximum length for each feature will be used. These
        lengths are computed while caching the datasets.
      num_examples: An optional maximum number of examples to take from the
        beginning of each Task dataset for evaluation.
      shuffle: Whether to shuffle the Task datasets. Only useful when
        `num_examples` is also set in order to get a semi-random subsample of
        the examples. Note that the shuffle will only be applied once during
        initialization (using `seed`) and the same subsample will be used on
        call to `evaluate`.
      logger_cls: A set of subclasses of `Logger` to write results with.
      log_dir: The directory to log outputs to. Required if `logger_cls` is
        non-empty.
      use_memory_cache: whether to use tf.data.Dataset#cache. May cause memory
        issues for large datasets.
      async_compute_metrics: Whether to compute metrics in a separate thread.
        Should be disabled if the metric evaluation uses the accelerator in
        order to avoid deadlocking (e.g., when being run alongside the training
        loop).
      target_field_name: Field name of the target in the input dataset examples.

    Raises:
      ValueError if `sequence_length` is None but a preprocessor depends on its
      value.
    """
    logging.info("Initializing Evaluator for '%s'", mixture_or_task_name)
    eval_tasks = dataset_providers.get_subtasks(
        dataset_providers.get_mixture_or_task(mixture_or_task_name)
    )
    self._eval_tasks = get_valid_eval_tasks(eval_tasks, eval_split)

    self._metrics_executor = (
        concurrent.futures.ThreadPoolExecutor(max_workers=1)
        if async_compute_metrics
        else None
    )
    self._metrics_future = None
    self._target_field_name = target_field_name
    self._cached_targets = collections.defaultdict(list)

    if not self._eval_tasks:
      logging.warning(
          "No eval task with valid split and metric fn found. Skipping eval."
      )
      return

    # Determine if sequence_length arg is required. This occurs when any of the
    # task preprocessors have a `sequence_length` arg with no default value.
    sequence_length_required = False
    for task in self._eval_tasks:
      for prep in task.preprocessors:
        prep_params = inspect.signature(prep).parameters
        if (
            "sequence_length" in prep_params
            and prep_params["sequence_length"].default
            == inspect.Parameter.empty
        ):
          if sequence_length is None:
            if isinstance(prep, functools.partial):
              prep_name = prep.func.__name__
            else:
              prep_name = prep.__name__
            raise ValueError(
                f"Preprocessor '{prep_name}' in task '{task.name}' has a "
                "`sequence_length` argument, making it incompatible with "
                "automatic sequence length detection. Pass a valid "
                "`sequence_length` to `Evaluator` and try again."
            )
          sequence_length_required = True
          break

    def dataset_fn(task: Task) -> tf.data.Dataset:
      return task.get_dataset(
          sequence_length=sequence_length,
          split=eval_split,
          shuffle=shuffle,
          num_epochs=1,
          seed=seed,
          use_cached=use_cached,
      )

    # `task_datasets` have the output features from seqio.Task.get_dataset.
    # These features will be converted to "model features" by the feature
    # converter before being cached.
    sequence_dims = {
        k: v.sequence_dim for k, v in feature_converter.TASK_FEATURES.items()
    }
    cached_task_datasets, max_lengths = _cache_and_measure_examples(
        tasks=self._eval_tasks,
        dataset_fn=dataset_fn,
        sequence_dims=sequence_dims,
        num_examples=num_examples,
        use_memory_cache=use_memory_cache,
    )

    if sequence_length is None:
      logging.info("Setting sequence lengths to %s", max_lengths)
      sequence_length = max_lengths
    else:
      log_long_warning = False
      log_same_warning = False

      sequence_length = {
          k: sequence_length.get(k, max_lengths[k]) for k in max_lengths
      }

      assert set(sequence_length.keys()) == set(
          max_lengths.keys()
      ), "sequence_length=%s limits must match the detected max_lengths=%s" % (
          sequence_length.keys(),
          max_lengths.keys(),
      )

      for k, l in sequence_length.items():
        if l is None:
          continue
        if isinstance(l, (tuple, list)):
          logging.warning(
              (
                  "Automatic length checking is not supported when lengths are"
                  "specified with a tuple for feature %s = %s. Please make "
                  "sure your max lengths are not removing parts of your inputs."
              ),
              k,
              l,
          )
        elif l > max_lengths[k]:
          log_long_warning = True
        elif not sequence_length_required and l == max_lengths[k]:
          log_same_warning = True

      if log_long_warning:
        logging.warning(
            (
                "Given sequence lengths are longer than necessary for some"
                " evaluation inputs or targets, resulting in wasted"
                " computation. Consider passing `None` for `sequence_length` to"
                " have them be automatically computed.\n Got: %s,\n Max"
                " Lengths: %s"
            ),
            sequence_length,
            max_lengths,
        )
      elif log_same_warning:
        logging.warning(
            (
                "Given sequence lengths *may be* insufficient for some"
                " evaluation inputs or targets. Such sequences will be"
                " truncated to fit, likely leading to sub-optimal results."
                " Consider passing `None` for `sequence_length` to have them be"
                " automatically computed.\n  Got: %s,\n Max Lengths: %s"
            ),
            sequence_length,
            max_lengths,
        )

    self._cached_model_datasets = {}

    if feature_converter.pack:
      raise ValueError("During evaluation, packing can't be used.")
    # Convert the task features to model features
    for task in self._eval_tasks:
      eval_ds = feature_converter(
          cached_task_datasets[task.name], sequence_length
      )

      # The eval dataset is enumerated to ensure that the order is preserved
      # throughout the entire evaluation process.
      self._cached_model_datasets[task.name] = eval_ds.enumerate()

    self._cached_task_datasets = cached_task_datasets
    self._model_feature_shapes = {
        k: tuple(spec.shape)
        for k, spec in eval_ds.element_spec.items()
        if spec.shape.rank > 0
    }

    if logger_cls and not log_dir:
      raise ValueError(
          "'log_dir' must be provided to `Evaluator` if `logger_cls` is "
          "non-empty."
      )
    if not logger_cls:
      logging.warning(
          "'logger_cls' is empty so seqio.Evaluator will not log its results."
      )
    self._loggers = tuple(cls(output_dir=log_dir) for cls in logger_cls)  # pytype:disable=not-instantiable

  def __del__(self):
    """Wait for metrics to be written before deletion."""
    if self._metrics_executor:
      self._metrics_executor.shutdown(wait=True)

  def close(self):
    """Wait for metrics to be written."""
    if self._metrics_executor:
      self._metrics_executor.shutdown(wait=True)

  def evaluate(
      self,
      *,
      compute_metrics: bool,
      step: Optional[int] = None,
      predict_fn: Optional[PredictFnCallable] = None,
      score_fn: Optional[ScoreFnCallable] = None,
      predict_with_aux_fn: Optional[PredictFnCallable] = None,
      model_fns: Optional[
          Mapping[metrics_lib.ModelOutputType, ModelFnCallable]
      ] = None,
  ) -> MetricsAndOutputsType:
    """Predict and score self.eval_tasks.

    Evaluation must preserve the example ordering. This requirement is satisfied
    by using enumerated dataset. Each of the cached eval task datasets is an
    enumerated tf.data.Dataset where each element has (index, example) format.
    Therefore, each index serves as a unique integer id for the example.

    `predict_fn` takes as input the cached eval dataset. The output
    may be of the form Sequence[(index, token_ids)] where `token_ids` is the
    sequence of token ids output by the model with the input `example` whose
    index matches `index`. Therefore, even if `predict_fn` mixes the order of
    the examples during prediction, the order can be corrected as long as the
    correct index for each example is maintained. `predict_with_aux_fn` is
    almost exactly the same as `predict_fn`, except that it also returns a
    dictionary of auxiliary values along with each sequence of `token_ids`.

    Similarly, `score_fn` takes the cached eval dataset as input and returns
    Sequence[(index, score)] where `score` is the sequence of log likelihood
    scores for the targets in the eval dataset.

    A common example is the multi-host setup where the evaluation dataset is
    split into multiple hosts that independently make predictions and combine
    the results during which the ordering can be mixed.

    There are 4 steps involved in the evaluation using predicted tokens:

    1. Model returns indices and output_tokens: Sequence[Tuple[int,
       Sequence[int]]], potentially with some auxiliary values.
    2. output tokens are decoded by `vocab.decode`
    3. Postprocessors are applied to the decoded output. These are denoted as
       predictions.
    4. Each metric function is applied to the predictions and the cached
       targets.

    Using auxiliary values is exactly the same as predicted tokens, except that
    a Mapping[str, Sequence[Any]] is also returned. Where len(Sequence[Any])
    should correspond to the number of elements in the dataset.

    There are 2 steps involved in the evaluation using scores:

    1. Model returns indices and scores: Sequence[Tuple[int, Sequence[float]]]
    2. Each metric function is applied to the scores and the cached targets.

    Args:
      compute_metrics: whether to compute metrics.
      step: an optional step number of the current evaluation. If unspecified, a
        dummy value of -1 will be used.
      predict_fn: a user-defined function, which takes in a tf.data.Dataset and
        outputs the sequence of predicted tokens. Only called if predict metrics
        exist for the tasks.
      score_fn: a user-defined function, which takes in a tf.data.Dataset and
        outputs the log likelihood score of the targets. Only called if score
        metrics exist for the task.
      predict_with_aux_fn: a user-defined function that has exactly the same
        behaviour as predict_fn, except that it also returns a dictionary of
        auxiliary values. Only called if predict_with_aux metrics exist for the
        tasks.
      model_fns: a dict mapping model output type to the model function
        (user-defined) that can produce outputs of that model output type.

    Returns:
      metrics: a Future containing a mapping from task name to computed metrics,
        or None if `compute_metrics` is False.
      all_output: a mapping from task name to all the output that metric
        evaluation needs.
    """
    # Reorganizes score_fn, predict_fn and predict_with_aux_fn into model_fns
    # for backward compatibility.
    model_fns = dict(model_fns or {})
    if predict_fn:
      model_fns[metrics_lib.ModelOutputType.PREDICTION] = predict_fn
    if score_fn:
      model_fns[metrics_lib.ModelOutputType.SCORE] = score_fn
    if predict_with_aux_fn:
      model_fns[metrics_lib.ModelOutputType.PREDICTION_WITH_AUX] = (
          predict_with_aux_fn
      )

    # Computes all the model outputs needed by metrics, and organizes them in
    # a dictionary structure - model_outputs: Mapping[ModelOutputType, Any]
    # make sure the examples are sorted by example index.
    all_output = {}

    for task in self.eval_tasks:
      logging.info("Evaluating %s", task.name)

      all_output[task.name] = {}
      # We loop over metrics and collect all the model outputs
      # that are needed for metric computation.
      for metric_obj in task.metric_objs:
        model_output_type = metric_obj.model_output_type
        if model_output_type not in all_output[task.name]:
          model_fn = model_fns[model_output_type]
          all_output[task.name][model_output_type] = _extract_model_output(
              self._cached_model_datasets[task.name], model_fn
          )

    if compute_metrics:
      if self._metrics_future:
        # Ensure previous step's metrics are finished and raise any exceptions
        # that may have occurred.
        tick = time.time()
        self._metrics_future.result()
        logging.info(
            "Time waiting for previous metrics run: %f secs.",
            time.time() - tick,
        )

      def compute_metrics_fn():
        tick = time.time()
        all_metrics: AllMetricsType = {}
        for task in self.eval_tasks:
          try:
            all_metrics[task.name] = self._compute_clu_metrics(
                task, all_output[task.name], step
            )
          except Exception as e:
            logging.error("Failed to evaluate task %s: %s", task.name, e)
            raise ValueError(f"Failed to evaluate task {task.name}: {e}") from e
        logging.info("Time computing metrics: %f secs.", time.time() - tick)
        return all_metrics

      def wrap_graph(fn):
        graph = tf.compat.v1.get_default_graph()

        def wrapped_fn():
          with graph.as_default():
            return fn()

        return wrapped_fn

      if not tf.executing_eagerly():
        compute_metrics_fn = wrap_graph(compute_metrics_fn)

      if self._metrics_executor:
        self._metrics_future = self._metrics_executor.submit(compute_metrics_fn)
      else:
        self._metrics_future = concurrent.futures.Future()
        self._metrics_future.set_result(compute_metrics_fn())
      all_metrics = self._metrics_future
    else:
      all_metrics = concurrent.futures.Future()
      all_metrics.set_result(None)
    return all_metrics, all_output

  def _compute_clu_metrics(
      self, task: Task, task_output, step: Optional[int] = None
  ) -> TaskMetricsType:
    """Computes and logs metrics given the predicted tokens and scores.

    Args:
      task: the task being evaluated.
      task_output: the model outputs needed by metrics of that task.
      step: an optional step number of the current evaluation. If unspecified, a
        dummy value of -1 will be used.

    Returns:
      The computed metrics.
    """
    logging.info("Computing metrics for %s", task.name)
    task_dataset = self.cached_task_datasets[task.name]

    task_metrics = []
    inferences = {}
    targets_and_inferences = None
    for metric_obj in task.metric_objs:
      model_output = task_output[metric_obj.model_output_type]
      # When model output type is PREDICTION_WITH_AUX or
      # SCORE_WITH_INTERMEDIATES, model output is a tuple of two arrays/lists.
      if isinstance(model_output, tuple):
        prediction_or_score, aux_value = model_output
        aux_value = jax.tree.map(
            np.array,
            aux_value,
            is_leaf=lambda x: isinstance(x, list),
        )
        model_output = (np.array(prediction_or_score), aux_value)
      else:
        model_output = np.array(model_output)
      metric_instance = metric_obj.from_model_output(
          tfds.as_numpy(task_dataset),
          model_output,
          task.output_features,
          self._target_field_name,
      )
      if isinstance(metric_instance, metrics_lib.CollectingMetric):
        metric_value, targets_and_inferences = metric_instance.actual_compute(
            tfds.as_numpy(task_dataset),
            task.output_features,
            self._target_field_name,
            self._cached_targets[task.name],
        )
        self._cached_targets[task.name] = targets_and_inferences["targets"]
      else:
        metric_value = metric_instance.compute()
        targets_and_inferences = None
        if hasattr(metric_instance, "targets_and_inferences"):
          targets_and_inferences = metric_instance.targets_and_inferences
      task_metrics.append(metric_value)
      # Records inferences for legacy logging compatibility.
      # common ones are score, output, prediction.
      if targets_and_inferences:
        for key, val in targets_and_inferences.items():
          if key == "targets":
            continue
          inferences[key] = val.tolist() if isinstance(val, np.ndarray) else val
    # Records targets for legacy logging compatibility.
    # Each targets_and_inferences should have identical targets.
    # Chooses the last targets_and_inferences for this recording purpose.
    if targets_and_inferences:
      targets = targets_and_inferences["targets"]
    else:
      targets = None

    result: TaskMetricsType = {}
    # TODO(b/309107492): Fix that attribute-error instead of silencing it.
    for k, v in itertools.chain(*[m.items() for m in task_metrics]):  # pytype: disable=attribute-error
      if k in result:
        raise ValueError(f"Duplicate metric key '{k}' in Task '{task.name}'.")
      result[k] = v

    metrics = {
        k: (
            metrics_lib.Scalar(v)
            if not isinstance(v, metrics_lib.MetricValue)
            else v
        )
        for k, v in result.items()
    }
    for logger in self.loggers:
      logger(
          task_name=task.name,
          step=step,
          metrics=metrics,
          dataset=task_dataset,
          targets=targets,
          inferences=inferences,
      )
    return result

  @property
  def eval_tasks(self) -> Sequence[Task]:
    return self._eval_tasks

  @property
  def cached_model_datasets(self) -> Mapping[str, tf.data.Dataset]:
    return self._cached_model_datasets

  @property
  def cached_task_datasets(self) -> Mapping[str, tf.data.Dataset]:
    return self._cached_task_datasets

  @property
  def model_feature_shapes(self) -> Mapping[str, Tuple[int, ...]]:
    return self._model_feature_shapes

  @property
  def loggers(self) -> Tuple[loggers_lib.Logger]:
    return tuple(self._loggers)


class MetricManager:
  """A class to manage metrics update and report for evaluation."""

  def __init__(self, tasks: Sequence[Task]):

    # metric_registry stores `Collection` (of metrics) class per task and
    # model_output_type.
    self.metric_registry = collections.defaultdict(dict)
    # output_metrics_collections stores `Collection` (of metrics) instance per
    # task and model_output_type.
    self.output_metrics_collections = collections.defaultdict(dict)
    for task in tasks:
      metric_objs = task.metric_objs

      metrics_per_output_type = collections.defaultdict(list)
      for metric_obj in metric_objs:
        model_output_type = metric_obj.model_output_type
        metrics_per_output_type[model_output_type].append(metric_obj)

      for model_output_type in metrics_per_output_type:
        metric_objs = metrics_per_output_type[model_output_type]
        metrics_collection = clu.metrics.Collection.create(**{
            type(metric).__name__ + f"_{idx}": type(metric)
            for idx, metric in enumerate(metric_objs)
        })
        self.metric_registry[task.name][model_output_type] = metrics_collection

  def initialize_metrics(
      self, task_name: str, model_output_type: ModelOutputType
  ):
    """Initializes metrics associated with the task name and model_output_type.

    This happens at the start of metric evaluation.

    Args:
      task_name: string name of the seqio task that we'd like to evlauate
        metrics on.
      model_output_type: we initialize metrics by the same model output type.
    """
    metrics_collection = self.metric_registry[task_name][model_output_type]
    self.output_metrics_collections[task_name][
        model_output_type
    ] = metrics_collection.empty()

  def from_model_output(
      self,
      task_name: str,
      model_output_type: ModelOutputType,
      inputs: Sequence[Mapping[str, Any]],
      model_output: Union[np.ndarray, Tuple[np.ndarray, np.ndarray]],
      features: Mapping[str, utils.Feature],
      target_field_name: str = "targets",
      mask: Optional[np.ndarray] = None,
      indices_2d: Optional[np.ndarray] = None,
  ):
    """Calculates the metrics associated with the given task name and model output type.

    Args:
      task_name: string name of the seqio task that we'd like to evlauate
        metrics on.
      model_output_type: we evaluate metrics by the same model output type.
      inputs: Examples in dataset.
      model_output: Model output computed by model functions.
      features: Output features defined in seqio.Task.
      target_field_name: Field name of the target sequence.
      mask: An array of booleans, same length as inputs. Each element indicates
        whether to include the corresponding element in the inputs for metric
        evaluation.
      indices_2d: 2d-indices of examples in the inputs/model_output. First
        dimension is shard id, the second is the example id within that shard.

    Returns:
    """
    metrics_collection = self.metric_registry[task_name][model_output_type]
    metric_batch = metrics_collection.single_from_model_output(
        inputs=inputs,
        model_output=model_output,
        indices_2d=indices_2d,
        features=features,
        target_field_name=target_field_name,
        mask=mask,
    )

    return metric_batch

  def merge(
      self,
      metrics_collection: clu.metrics.Collection,
      task_name: str,
      model_output_type: ModelOutputType,
  ):
    """Updates the metrics for the given task and model output type.

    Args:
      metrics_collection: A clu metric Collection object. Usually it is
        evaluated on a batch.
      task_name: string name of the seqio task that we'd like to evlauate
        metrics on.
      model_output_type: we evaluate metrics by the same model output type.
    """
    curr_metrics_collection = self.output_metrics_collections[task_name][
        model_output_type
    ]
    new_metrics_collection = curr_metrics_collection.merge(metrics_collection)
    self.output_metrics_collections[task_name][
        model_output_type
    ] = new_metrics_collection

  def host_gather_and_reduce(
      self, task_name: str, model_output_type: ModelOutputType
  ):
    """Gathers metrics across hosts for the given task and model output type.

    Args:
      task_name: string name of the seqio task that we'd like to evlauate
        metrics on.
      model_output_type: we evaluate metrics by the same model output type.
    """
    curr_metrics_collection = self.output_metrics_collections[task_name][
        model_output_type
    ]
    gathered = multihost_utils.process_allgather(curr_metrics_collection)
    self.output_metrics_collections[task_name][
        model_output_type
    ] = gathered.reduce()



# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for task_registry_provenance_tracking."""

import inspect

from absl.testing import absltest
from seqio import dataset_providers
from seqio import task_registry_provenance_tracking


class TestRegistryProvenanceTrackingTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    task_registry_provenance_tracking.turn_on_tracking()

  def tearDown(self):
    super().tearDown()
    task_registry_provenance_tracking._PROVIDER_PROVENANCE_LOOKUP = None

  def test_track_registration_provenance(self):
    input_task_name = "test_task_track_creation_provenance"

    input_data_source = dataset_providers.TextLineDataSource({})
    input_output_features = {}

    # Use a closure just to test the "more difficult" case for tracking
    # provenance of having extra closed-over state.
    def _closure_containing_task_init():
      task = dataset_providers.Task(
          input_task_name,
          input_data_source,
          input_output_features,
      )

      # These two statements are on the same line because the registration
      # provenance tracking looks at filename and line number; in order to get a
      # correct expected value, both things need to happen on the same line.
      # pyformat: disable
      expected_registration_provenance = task_registry_provenance_tracking.informative_upstream_callsites_from_frame(inspect.currentframe()); dataset_providers.TaskRegistry.add_provider(input_task_name, task)  # pylint: disable=multiple-statements
      # pyformat: enable
      return expected_registration_provenance

    expected_provider_type = dataset_providers.Task.__name__
    expected_registration_provenance = _closure_containing_task_init()

    actual_provider_type, actual_registration_provenance = (
        task_registry_provenance_tracking.view_provenances()[input_task_name]
    )

    self.assertEqual(actual_provider_type, expected_provider_type)

    self.assertEqual(
        actual_registration_provenance,
        expected_registration_provenance,
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for data loading and processing."""

import collections
import contextlib
import dataclasses
import functools
import inspect
import os
import re
import types
from typing import Any, Callable, Dict, Iterable, Mapping, Optional, Sequence, Tuple, Type, Union

from absl import logging
import numpy as np
from seqio.vocabularies import Vocabulary
import tensorflow.compat.v2 as tf
import tensorflow_datasets as tfds

_INFO_FILENAME = "info.{split}.json"
_STATS_FILENAME = "stats.{split}.json"
_TFRECORD_PREFIX = "{split}.tfrecord"

_TFDS_DATA_DIR_OVERRIDE = None
_TFDS_DATA_READ_CONFIG_OVERRIDE = None
_GLOBAL_CACHE_DIRECTORIES = []
_MapTransform = object
_RandomMapTransform = object


@dataclasses.dataclass(frozen=True)
class Feature:
  """A container for attributes of output features of data providers."""

  vocabulary: Vocabulary
  add_eos: bool = True
  required: bool = True
  dtype: tf.DType = tf.int32
  rank: int = 1



def set_tfds_data_dir_override(tfds_data_dir):
  global _TFDS_DATA_DIR_OVERRIDE
  _TFDS_DATA_DIR_OVERRIDE = tfds_data_dir




def set_tfds_read_config_override(tfds_read_config):
  global _TFDS_DATA_READ_CONFIG_OVERRIDE
  _TFDS_DATA_READ_CONFIG_OVERRIDE = tfds_read_config




def get_global_cache_dirs():
  return _GLOBAL_CACHE_DIRECTORIES


def set_global_cache_dirs(global_cache_dirs):
  global _GLOBAL_CACHE_DIRECTORIES
  _GLOBAL_CACHE_DIRECTORIES = []
  add_global_cache_dirs(global_cache_dirs)


def add_global_cache_dirs(global_cache_dirs):
  for cache_dir in global_cache_dirs:
    if cache_dir not in _GLOBAL_CACHE_DIRECTORIES:
      _GLOBAL_CACHE_DIRECTORIES.append(cache_dir)




def _validate_tfds_name(name: str) -> None:
  """Validates TFDS dataset name."""
  if (
      name
      and ":" not in name
  ):
    raise ValueError(f"TFDS name must contain a version number, got: {name}")


def _get_data_dir_override(tfds_name: Optional[str]) -> Optional[str]:
  """Returns the data dir in case it is overridden."""
  if (
      _TFDS_DATA_DIR_OVERRIDE
  ):
    return _TFDS_DATA_DIR_OVERRIDE
  return None


@dataclasses.dataclass(frozen=True)
class TfdsSplit:
  """Points to a specific TFDS split.

  Attributes:
    dataset: dataset name.
    split: TFDS split (e.g. 'train'), or slice (e.g. 'train[":1%"]').
    data_dir: directory to read/write TFDS data.
  """

  dataset: str
  split: Optional[str]
  data_dir: Optional[str] = None

  def __post_init__(self):
    _validate_tfds_name(self.dataset)


class LazyTfdsLoader(object):
  """Wrapper for TFDS datasets with memoization and additional functionality.

  Lazily loads info from TFDS and provides memoization to avoid expensive hidden
  file operations. Also provides additional utility methods.
  """

  _MEMOIZED_BUILDERS = {}

  def __init__(
      self,
      name: Optional[str] = None,
      data_dir: Optional[str] = None,
      split_map: Union[Mapping[str, str], Mapping[str, TfdsSplit], None] = None,
      decoders=None,
      builder_kwargs: Optional[dict[str, Any]] = None,
      read_only: bool = False,
  ):
    """LazyTfdsLoader constructor.

    Args:
      name: str (optional), the name of the TFDS dataset. If `name` is not
        specified then `split_map` values must be instances of `TfdsSplit`.
      data_dir: str (optional), directory to read/write TFDS data.
      split_map: dict (optional), mapping from canonical splits (e.g.,
        'validation') to TFDS splits (e.g. 'train'), or slices (e.g.,
        'train[':1%']), or `TfdsSplit` (e.g. `TfdsSplit(dataset='mnist',
        split='train')`). If `TfdsSplit` are used then `name` must be empty.
      decoders: dict (optional), mapping from features to tfds.decode.Decoders,
        such as tfds.decode.SkipDecoding() for skipping image byte decoding.
      builder_kwargs: `dict` (optional), keyword arguments to be passed to the
        `tfds.core.DatasetBuilder` constructor through `tfds.load()` and
        `tfds.builder()`.
      read_only: whether `get_dataset` can trigger the generation of a dataset.
    """
    _validate_tfds_name(name)
    self._name = name
    self._data_dir = data_dir
    self._data_dir_override = None
    self._split_map = split_map
    self._decoders = decoders
    self._builder_kwargs = builder_kwargs
    self._read_only = read_only

    self._is_custom_split_map = False
    if split_map:
      random_split_value = next(iter(split_map.values()))
      if isinstance(random_split_value, TfdsSplit):
        self._is_custom_split_map = True
        if self._name or self._data_dir:
          raise ValueError(
              "If split values are instances of `TfdsSplit`, `name` and"
              " `data_dir` must be `None`."
          )

  @property
  def name(self) -> Optional[str]:
    return self._name

  def set_decoders(self, decoders) -> None:
    """Override the decoders for this dataset."""
    self._decoders = decoders

  @property
  def tfds_splits(self) -> Optional[Mapping[str, TfdsSplit]]:
    return self._split_map if self._is_custom_split_map else None

  def resolved_tfds_name(self, split: Optional[str] = None) -> Optional[str]:
    """Returns the resolved TFDS dataset name.

    When the specified TFDS name doesn't specify everything, e.g. the version
    has a wildcard or the config is not specified, then this function returns
    the complete TFDS name if the dataset has already been loaded.

    Args:
      split: optional split name.

    Returns:
      complete TFDS name.
    """
    if self.is_memoized(split):
      return (
          self._get_builder(split)
          .get_reference()
          .tfds_name(include_version=True)
      )
    else:
      dataset, _ = self.get_split_params(split)
      return dataset

  def __str__(self):
    return (
        f"{self.__class__.__name__}(name={self.name}, data_dir={self.data_dir})"
    )

  def __repr__(self):
    return (
        f"{self.__class__.__name__}("
        f"name={self.name},"
        f" data_dir={self.data_dir},"
        f" split_map={self._split_map},"
        f" decoders={self._decoders})"
    )

  def get_split_params(
      self, split: Optional[str] = None
  ) -> Tuple[Optional[str], Optional[str]]:
    """Returns a tuple of (dataset, data_dir) for the given canonical split."""
    if self._is_custom_split_map:
      if mapped_split := self._split_map.get(split):
        dataset = mapped_split.dataset
        data_dir = mapped_split.data_dir
      else:
        raise ValueError(
            "`LazyTfdsLoader` refers to multiple datasets, pass `split` value "
            "corresponding to one of them to `get_split_params()`."
        )
    else:
      dataset = self.name
      data_dir = self.data_dir

    return dataset, data_dir


  @property
  def data_dir(self) -> Optional[str]:
    """Returns the data directory for this TFDS dataset."""

    if self._is_custom_split_map:
      logging.warning(
          "`LazyTfdsLoader` refers to multiple datasets, `data_dir` is unknown."
      )
      return None

    if self._data_dir_override is None:
      if data_dir_override := _get_data_dir_override(tfds_name=self.name):
        self._data_dir_override = data_dir_override

    if self._data_dir_override and self._data_dir:
      logging.warning(
          "Overriding TFDS data directory '%s' with '%s' for dataset '%s'.",
          self._data_dir,
          self._data_dir_override,
          self.name,
      )

    return self._data_dir_override or self._data_dir

  def override_data_dir(self, data_dir: str) -> None:
    self._data_dir_override = data_dir

  @property
  def read_config(self):
    if _TFDS_DATA_READ_CONFIG_OVERRIDE:
      return _TFDS_DATA_READ_CONFIG_OVERRIDE
    return tfds.ReadConfig()

  def _get_builder_key(
      self, dataset: Optional[str], data_dir: Optional[str]
  ) -> Tuple[Optional[str], Optional[str]]:
    return (dataset, data_dir)

  def is_memoized(self, split: Optional[str] = None) -> bool:
    """Returns true if the dataset is memoized."""
    dataset, data_dir = self.get_split_params(split)

    return (
        self._get_builder_key(dataset, data_dir)
        in LazyTfdsLoader._MEMOIZED_BUILDERS
    )

  def builder_cls(
      self, split: Optional[str] = None
  ) -> Optional[Type[tfds.core.DatasetBuilder]]:
    """Returns the DatasetBuilder class for this TFDS dataset.

    If no builder class can be found (e.g. the class has not been imported),
    then `None` will be returned.

    Args:
      split: optional split for which to return the builder class. This can be
        used in case there's a custom split map with different datasets per
        split.

    Returns:
      the builder class of the dataset in case it can be found and `None` if the
      builder class cannot be found.
    """
    dataset, _ = self.get_split_params(split)
    # `builder_cls` only accepts the dataset name, not the config and version.
    ds_name, _ = tfds.core.naming.parse_builder_name_kwargs(dataset)
    try:
      return tfds.builder_cls(str(ds_name))
    except ValueError:
      return None

  @property
  def builder(self):
    return self._get_builder()

  def _get_builder(self, split: Optional[str] = None):
    """Returns the DatasetBuilder for this TFDS dataset."""
    dataset, data_dir = self.get_split_params(split)
    builder_key = self._get_builder_key(dataset, data_dir)
    if builder_key not in LazyTfdsLoader._MEMOIZED_BUILDERS:
      if dataset:
        builder_kwargs = self._builder_kwargs if self._builder_kwargs else {}
        builder = tfds.builder(dataset, data_dir=data_dir, **builder_kwargs)
      else:
        if self._builder_kwargs:
          raise ValueError(
              "`builder_kwargs` should be empty when `dataset` value is not"
              " present."
          )
        builder = tfds.builder_from_directory(data_dir)
      LazyTfdsLoader._MEMOIZED_BUILDERS[builder_key] = builder
    return LazyTfdsLoader._MEMOIZED_BUILDERS[builder_key]

  @property
  def info(self):
    return self.builder.info

  def _map_split(self, split: str) -> Optional[str]:
    """Maps the given split to a dataset split."""
    if self._is_custom_split_map:
      self._split_map: Mapping[str, TfdsSplit]
      return self._split_map[split].split
    elif self._split_map:
      self._split_map: Mapping[str, str]
      return self._split_map[split]
    else:
      return split

  def files(self, split: str):
    """Returns set of instructions for reading TFDS files for the dataset."""
    dataset_split = self._map_split(split)
    builder = self._get_builder(split)

    if (
        self.name is not None
        and "/" not in self.name
        and builder.BUILDER_CONFIGS
    ):
      # If builder has multiple configs, and no particular config was
      # requested, raise an error.
      raise ValueError("Dataset '%s' has multiple configs." % self.name)

    split_info = builder.info.splits[dataset_split]
    files = split_info.file_instructions

    if not files:
      logging.fatal("No TFRecord files found for dataset: %s", self.name)
    return files

  def load(
      self,
      split: Optional[str],
      shuffle_files: bool,
      seed: Optional[int] = None,
      shard_info=None,
  ):
    """Returns a tf.data.Dataset for the given split."""
    dataset_split = self._map_split(split)
    dataset, data_dir = self.get_split_params(split)
    read_config = self.read_config
    read_config.input_context = (
        tf.distribute.InputContext(  # pylint: disable=g-long-ternary
            num_input_pipelines=shard_info.num_shards,
            input_pipeline_id=shard_info.index,
        )
        if shard_info
        else None
    )
    read_config.shuffle_seed = seed
    read_config.skip_prefetch = True
    if self._read_only:
      return self.builder.as_dataset(
          split=dataset_split,
          shuffle_files=shuffle_files,
          read_config=read_config,
          decoders=self._decoders,
      )
    else:
      return tfds.load(
          dataset,
          split=dataset_split,
          data_dir=data_dir,
          shuffle_files=shuffle_files,
          download=True,
          try_gcs=True,
          read_config=read_config,
          decoders=self._decoders,
          builder_kwargs=self._builder_kwargs,
      )

  def load_shard(
      self,
      file_instruction,
      shuffle_files: bool = False,
      seed: Optional[int] = None,
  ):
    """Returns a dataset for a single shard of the TFDS TFRecord files."""
    # pytype:disable=attribute-error
    ds = self.builder._tfrecords_reader.read_files(  # pylint:disable=protected-access
        [file_instruction],
        read_config=tfds.ReadConfig(shuffle_seed=seed),
        shuffle_files=shuffle_files,
    )
    # pytype:enable=attribute-error
    return ds

  def size(self, split: str) -> Optional[int]:
    """Returns the number of examples in the split."""
    dataset_split = self._map_split(split)
    ds_splits = self._get_builder(split).info.splits
    dataset_size = ds_splits[dataset_split].num_examples
    # Very large datasets have num_examples = 0; default instead to np.inf
    dataset_size = dataset_size if dataset_size > 0 else np.inf
    return dataset_size


# ============================== TFExamples ====================================

# Type alias for "dictionary of tensors"
TFDict = Dict[
    str, Union[np.ndarray, tf.Tensor, tf.RaggedTensor, tf.SparseTensor]
]
# NOTE: We use short prefixes to minimize feature key overhead.
# Used to demarcate features used to store shapes of other tensor
_TFEXAMPLE_SHAPE_PREFIX = "_sh:"
# See `tfexample_ragged_prefix`
_TFEXAMPLE_RAGGED_PREFIX = "_rl:"
# See `tfexample_sparse_indices_prefix`
_TFEXAMPLE_SPARSE_PREFIX = "_sp:"


def tfexample_ragged_length_key(key: str, dim: int) -> str:
  """Demarcates feature used to store `dim`-th ragged lengths for `key`.

  This function can be used when parsing data generated by `dict_to_tfexample`,
  by specifying the following ragged feature:
  >>> KEY = "ragged_tensor"
  >>> tf.io.parse_single_example(example, {
  >>>   KEY: tf.io.RaggedFeature(
  >>>     dtype,
  >>>     value_key=KEY,
  >>>     partitions=(
  >>>       tf.io.RaggedFeature.RowLengths(tfexample_ragged_length_key(KEY, 0)),
  >>>       tf.io.RaggedFeature.RowLengths(tfexample_ragged_length_key(KEY, 1)),
  >>>       ... (repeat for however many ragged dimensions the data has)
  >>>   ),)
  >>> })

  Args:
    key: The key storing the values of the ragged tensor.
    dim: The ragged dimension to generate a prefix for.

  Returns:
    The key of the feature storing the `dim`-th ragged lengths.
  """
  return f"{_TFEXAMPLE_RAGGED_PREFIX}{dim}:{key}"


def tfexample_sparse_indices_key(key: str, dim: int) -> str:
  """Demarcates feature used to store `dim`-th sparse feature indices for `key`.

  This function can be used when parsing data generated by `dict_to_tfexample`,
  by specifying the following sparse feature:
  >>> KEY = "sparse_tensor"
  >>> tf.io.parse_single_example(example, {
  >>>   KEY: tf.io.SparseFeature(
  >>>     value_key=KEY,
  >>>     index_key=[
  >>>       tfexample_sparse_indices_key(KEY, 0),
  >>>       tfexample_sparse_indices_key(KEY, 1),
  >>>       ... (repeat for however many sparse dimensions the data has)
  >>>     ],
  >>>     size=[shape0, shape1, ...],
  >>>     dtype=dtype,
  >>>   ),
  >>> })

  Args:
    key: The key storing the values of the sparse tensor.
    dim: The sparse dimension to generate a prefix for.

  Returns:
    The key of the feature storing the `dim`-th sparse indices.
  """
  return f"{_TFEXAMPLE_SPARSE_PREFIX}{dim}:{key}"


def _to_tffeature(tensor: tf.Tensor) -> tf.train.Feature:
  """Creates an appropriately typed `tf.train.Feature` for `tensor`.

  This method only stores the flattened values of `tensor`; make sure you have
  stored the shape before use.

  Args:
    tensor: A tensor to convert.

  Returns:
    A `tf.train.Feature` for `tensor`
  """
  # Flatten the values of the  tensor.
  values = tf.reshape(tensor, [-1]).numpy().tolist()

  if tensor.dtype.is_bool or tensor.dtype.is_integer:
    return tf.train.Feature(int64_list=tf.train.Int64List(value=values))
  elif tensor.dtype.is_floating:
    return tf.train.Feature(float_list=tf.train.FloatList(value=values))
  elif tensor.dtype is tf.string:
    values = [tf.compat.as_bytes(v) for v in values]
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=values))
  else:
    raise ValueError(f"Unsupported type {tensor.dtype}")


def dict_to_tfexample(
    dct: TFDict, store_shapes: bool = False
) -> tf.train.Example:
  """Convert dictionary of tensors to a `tf.train.Example` proto.

  NOTE: Unfortunately, tensorflow.Example is a very simple proto that can only
  store keyed lists of int64, float and bytes, and doesn't have a marker to
  store metadata needed for multi-dimensional or ragged tensors. This function
  stores using additional features; using `tfexample_to_dict` will allow you to
  recover the original dictionary of tensors modulo the following caveat.

  tensorflow.train.Example only stores int64s and float32s: this function
  casts bools, ints and floats respectively, and this type information is lost.

  Args:
    dct: A dictionary mapping feature keys to 1-D tensors.
    store_shapes: If true, add an additional feature to store the shapes of any
      2+d or sparse tensors. This feature is only required when using
      `tfexample_to_dict`, as `tf.io.parse_single_example` requires the shape be
      provided as an argument.

  Returns:
    A `tf.train.Example` for `dct`
  """
  features = {}
  for key, value in dct.items():
    if isinstance(value, tf.RaggedTensor):
      value: tf.RaggedTensor
      features[key] = _to_tffeature(value.flat_values)
      for dim, length in enumerate(value.nested_row_lengths()):
        features[tfexample_ragged_length_key(key, dim)] = _to_tffeature(length)
    elif isinstance(value, tf.SparseTensor):
      value: tf.SparseTensor
      features[key] = _to_tffeature(value.values)
      for dim, indices in enumerate(tf.transpose(value.indices)):
        features[tfexample_sparse_indices_key(key, dim)] = _to_tffeature(
            indices
        )
      if store_shapes:
        features[_TFEXAMPLE_SHAPE_PREFIX + key] = _to_tffeature(
            tf.constant(value.shape)
        )
    else:
      # Cast to a dense tensor.
      value = tf.constant(value)
      if store_shapes and len(value.shape) > 1:
        features[_TFEXAMPLE_SHAPE_PREFIX + key] = _to_tffeature(
            tf.constant(value.shape)
        )
      features[key] = _to_tffeature(value)

  return tf.train.Example(features=tf.train.Features(feature=features))


def tfexample_to_dict(example: tf.train.Example) -> TFDict:
  """Helper function to create a dictionary of tensors from a TFExample.

  NOTE: this function is less efficient than `tf.io.parse_single_example`, but
  allows parsing of TFExample without knowing its features ahead of time. See
  the documentation of `tfexample_ragged_length_key` for an example of how to
  parse ragged tensors efficiently using `tf.io.parse_single_example`.

  Args:
    example: An instance of tf.train.Example we will convert into a dict.

  Returns:
    A dict with the keys of `example` and tensors as values.
  """
  dct = {}
  shapes = {}
  ragged_lengths = collections.defaultdict(dict)
  sparse_indices = collections.defaultdict(dict)

  for key, feature in example.features.feature.items():
    if feature.int64_list.value:
      value = tf.constant(list(feature.int64_list.value), dtype=tf.int64)
    elif feature.float_list.value:
      value = tf.constant(list(feature.float_list.value), dtype=tf.float32)
    elif feature.bytes_list.value:
      value = tf.constant(list(feature.bytes_list.value), dtype=tf.string)
    else:
      # This is an empty list. We don't know what the type is, so we default
      # to tf.string
      value = tf.constant(list(feature.bytes_list.value), dtype=tf.string)

    if key.startswith(_TFEXAMPLE_RAGGED_PREFIX):
      _, dim, key = key.split(":", 2)
      ragged_lengths[key][int(dim)] = value
    elif key.startswith(_TFEXAMPLE_SPARSE_PREFIX):
      _, dim, key = key.split(":", 2)
      sparse_indices[key][int(dim)] = value
    elif key.startswith(_TFEXAMPLE_SHAPE_PREFIX):
      key = key[len(_TFEXAMPLE_SHAPE_PREFIX) :]
      shapes[key] = value
    else:
      dct[key] = value

  # Assemble RaggedTensors
  for key, length_map in ragged_lengths.items():
    flat_values = dct[key]
    nested_row_lengths = []
    for dim in range(len(length_map)):
      if dim not in length_map:
        raise ValueError(f"Couldn't find {dim}-th ragged length for {key}")
      nested_row_lengths.append(length_map[dim])
    dct[key] = tf.RaggedTensor.from_nested_row_lengths(
        flat_values, nested_row_lengths
    )

  # Assemble SparseTensors
  for key, indices_map in sparse_indices.items():
    if key not in shapes:
      raise ValueError(f"Couldn't find dense shape for sparse tensor {key}")
    dense_shape = shapes.pop(key)

    values = dct[key]
    indices_per_dim = []
    for dim in range(len(indices_map)):
      if dim not in indices_map:
        raise ValueError(f"Couldn't find {dim}-th sparse indices for {key}")
      indices_per_dim.append(indices_map[dim])
    indices = tf.transpose(tf.stack(indices_per_dim))
    dct[key] = tf.SparseTensor(indices, values, dense_shape)

  # Reshape multi-dimensional tensors
  for key, shape in shapes.items():
    dct[key] = tf.reshape(dct[key], shape)
  return dct


# Type alias that supports nested TFDicts.
NestedTFDict = Dict[
    str, Union[tf.Tensor, tf.RaggedTensor, tf.SparseTensor, "NestedTFDict"]
]
# Used to linearize keys in nested TFDicts
_TFEXAMPLE_NESTED_DELIMITER = "/"


def unflatten_dict(
    dct: TFDict, delimiter=_TFEXAMPLE_NESTED_DELIMITER
) -> NestedTFDict:
  """Create a nested dictionary from one with nested keys.

  This method converts a "flat" TFDict with nested keys like:
  >>> {
  >>>     "key1/subkey1": ...,
  >>>     "key1/subkey2": ...,
  >>>     "key2/subkey1": ...,
  >>> }
  into a nested dictionary:
  >>> {
  >>>     "key1": {
  >>>       "subkey1": ...,
  >>>       "subkey2": ...,
  >>>     },
  >>>     "key2": {
  >>>       "subkey3": ...
  >>>     },
  >>> }

  Args:
    dct: An dictionary of tensors.
    delimiter: A delimiter used to separate keys from subkeys.

  Returns:
    A nested-version of `dct`.
  """
  nested_dct: NestedTFDict = {}

  for key_path, value in dct.items():
    keys = key_path.split(delimiter)
    # We'll index the value at the last key.
    last_key = keys.pop()

    sub_dct = nested_dct
    for key in keys:
      sub_dct = sub_dct.setdefault(key, {})
    sub_dct[last_key] = value
  return nested_dct


def flatten_dict(
    nested_dct: NestedTFDict, delimiter=_TFEXAMPLE_NESTED_DELIMITER
) -> TFDict:
  """Create a "flattened" dictionary from one with nested keys.

  This method converts a nested TFDict like:
  >>> {
  >>>     "key1": {
  >>>       "subkey1": ...,
  >>>       "subkey2": ...,
  >>>     },
  >>>     "key2": {
  >>>       "subkey1": ...
  >>>     },
  >>> }
  into a flattened dictionary:
  >>> {
  >>>     "key1/subkey1": ...,
  >>>     "key1/subkey2": ...,
  >>>     "key2/subkey1": ...,
  >>> }

  Args:
    nested_dct: A nested dictionary of tensors.
    delimiter: A delimiter used to separate keys from subkeys.

  Returns:
    A flattened version of `nested_dct`.
  """
  unnested_dct = {}

  def _unnest_dct(dct: NestedTFDict, prefix_key: str = ""):
    for key, value in dct.items():
      key_ = prefix_key + key
      if isinstance(value, dict):
        _unnest_dct(value, prefix_key + key + delimiter)
      else:
        unnested_dct[key_] = value

  _unnest_dct(nested_dct)
  return unnested_dct


# ================================ Tasks =======================================


def add_kwargs_to_transform(transform, **kwargs):
  """Returns the partial function or dataclass with the kwargs.

  We use this function to add common arguments (sequence_length,
  output_features) on all transformations that require those.

  Args:
    transform: A dataclass or function.
    **kwargs: Arguments to be passed to the transform.

  Returns:
    If `transform` is a dataclasses attributes matching the kwargs keys will be
    set to the kwargs values.
    Otherwise if `transform` is a function and takes any of the provided kwargs
    these will be passed (by running a partial function).
  """
  is_dataclass = dataclasses.is_dataclass(transform)
  # Filter kwargs by attributes of the dataclass/arguments of the function.
  if is_dataclass:
    available_arg_names = [f.name for f in dataclasses.fields(transform)]
  else:
    available_arg_names = set(inspect.signature(transform).parameters.keys())
    if isinstance(transform, functools.partial):
      available_arg_names -= set(transform.keywords.keys())
  kwargs = {k: v for k, v in kwargs.items() if k in available_arg_names}
  if not kwargs:
    return transform
  # Add attributes/arguments.
  if is_dataclass:
    return dataclasses.replace(transform, **kwargs)
  return functools.partial(transform, **kwargs)


def get_cached_info_path(data_dir, split):
  return os.path.join(data_dir, _INFO_FILENAME.format(split=split))


def get_cached_tfrecord_prefix(data_dir, split):
  return os.path.join(data_dir, _TFRECORD_PREFIX.format(split=split))


def get_cached_stats_path(data_dir, split):
  return os.path.join(data_dir, _STATS_FILENAME.format(split=split))


def get_task_dir_from_name(task_name):
  return os.path.join(*(task_name.split(":")))


def stateless_shuffle(value, seed):
  """Randomly shuffles a tensor, statelessly."""
  flat_value = tf.reshape(value, [-1])
  indices = tf.argsort(
      tf.random.stateless_uniform(tf.shape(flat_value), seed=seed)
  )
  flat_shuffle = tf.gather(flat_value, indices)
  return tf.reshape(flat_shuffle, tf.shape(value))


def trim_and_pad_dataset(
    dataset: tf.data.Dataset, feature_lengths: Mapping[str, int]
) -> tf.data.Dataset:
  """Trim and pad first dimension of features to `feature_lengths`.

  Args:
    dataset: tf.data.Dataset, the dataset to trim/pad examples in.
    feature_lengths: map from feature key to final length. Other features will
      be returned unchanged.

  Returns:
    Trimmed/padded tf.data.Dataset.
  """

  def _trim_and_pad(k: str, t: tf.Tensor) -> tf.Tensor:
    """Trim/pad to the first axis of `t` to be of size `length`."""
    if k not in feature_lengths:
      return t
    if isinstance(t, tf.RaggedTensor):
      t = t.to_tensor()

    length_k = feature_lengths[k]
    if isinstance(length_k, int):
      t = t[:length_k]
      pad_amt = length_k - tf.shape(t)[0]
      padded_t = tf.pad(t, [(0, pad_amt)] + [(0, 0)] * (len(t.shape) - 1))
      padded_t.set_shape([length_k] + t.shape.as_list()[1:])
      return padded_t

    slices = tuple((slice(0, limit) for limit in length_k))
    t = t[slices]
    pad_amt = tf.pad((length_k - tf.shape(t))[..., None], ((0, 0), (1, 0)))
    padded_t = tf.pad(t, pad_amt)
    padded_t.set_shape(length_k)
    return padded_t

  return dataset.map(
      lambda x: {k: _trim_and_pad(k, t) for k, t in x.items()},
      num_parallel_calls=tf.data.experimental.AUTOTUNE,
  )


def trim_dataset(
    dataset: tf.data.Dataset, sequence_length, output_features
) -> tf.data.Dataset:
  """Trim output features to sequence length."""

  def _trim(k: str, v: tf.Tensor) -> tf.Tensor:
    if (
        k not in output_features
        or not sequence_length
        or k not in sequence_length
        or sequence_length[k] is None
    ):
      return v
    # Unify lengths into an iterable so we can create a slice for each
    # dimension, even if the length is a single int.
    lengths = sequence_length[k]
    if isinstance(lengths, int):
      lengths = [lengths]
    slices = tuple((slice(0, limit) for limit in lengths))
    return v[slices]

  return dataset.map(
      lambda ex: {k: _trim(k, v) for k, v in ex.items()},
      num_parallel_calls=tf.data.experimental.AUTOTUNE,
  )


def _strip_packed_feature_key(key: str) -> str:
  strip_suffix = lambda k, s: k[: -len(s)] if k.endswith(s) else k
  return strip_suffix(strip_suffix(key, "_positions"), "_segment_ids")


def trim_and_pack_dataset(
    dataset: tf.data.Dataset,
    feature_lengths: Mapping[str, int],
    use_custom_ops: bool = False,
) -> tf.data.Dataset:
  """Creates a 'packed' version of a dataset on-the-fly.

  Modified from the tensor2tensor library.

  This is meant to replace the irritation of having to create a separate
  "packed" version of a dataset to train efficiently on TPU.

  Each example in the output dataset represents several examples in the
  input dataset.

  For each key in the input dataset that also exists in `feature_lengths`, two
  additional keys are created:
    <key>_segment_ids: an int32 tensor identifying the parts
       representing the original example.
    <key>_positions: an int32 tensor identifying the position within the
       original example.

  Features that are not in `feature_lengths` will be removed.

  Example:
    Two input examples get combined to form an output example.
    The input examples are:
    {"inputs": [8, 7, 1, 0], "targets":[4, 1, 0], "idx": 0}
    {"inputs": [2, 3, 4, 1], "targets":[5, 6, 1], "idx": 1}
    The output example is:
    {
                   "inputs": [8, 7, 1, 2, 3, 4, 1, 0, 0, 0]
       "inputs_segment_ids": [1, 1, 1, 2, 2, 2, 2, 0, 0, 0]
         "inputs_positions": [0, 1, 2, 0, 1, 2, 3, 0, 0, 0]
                  "targets": [4, 1, 5, 6, 1, 0, 0, 0, 0, 0]
      "targets_segment_ids": [1, 1, 2, 2, 2, 0, 0, 0, 0, 0]
        "targets_positions": [0, 1, 0, 1, 2, 0, 0, 0, 0, 0]
    }

    0 represents padding in both the inputs and the outputs.

    Sequences in the incoming examples are truncated to length in
    `feature_lengths`, and the sequences in the output examples all have this
    fixed (padded) length. Features not in `features_length` (i.e, "idx") are
    removed.

  Args:
    dataset: a tf.data.Dataset
    feature_lengths: map from feature key to final length. Other features will
      be discarded.
    use_custom_ops: a boolean - custom ops are faster but require a custom-built
      binary, which is not currently possible on cloud-tpu.

  Returns:
    a tf.data.Dataset
  """
  element_spec = dataset.element_spec
  # Make sure that the dataset contains all keys in `feature_lengths`.
  for k in feature_lengths:
    if k not in element_spec:
      raise ValueError(
          f"Feature '{k}' not found in dataset. Available keys are "
          f"{list(element_spec.keys())}"
      )
    if (
        not element_spec[k].shape.is_compatible_with(tf.TensorShape([None]))
        and not use_custom_ops
    ):
      raise ValueError(
          f"Features to be packed must be one-dimensional. '{k}' is not.' "
          "Consider setting use_custom_ops if you have higher-rank features."
      )

  # Warn if there are any additional keys that will be removed.
  additional_keys = set(element_spec) - set(feature_lengths)
  if additional_keys:
    logging.warning(
        "Features not in `features_length` will be removed during packing: %s",
        additional_keys,
    )

  ds = dataset.map(
      lambda x: {k: x[k][:l, ...] for k, l in feature_lengths.items()},
      num_parallel_calls=tf.data.experimental.AUTOTUNE,
  )

  # Setting batch_size=length ensures that the concatenated sequences (if they
  # have length >=1) are sufficient to fill at least one packed example.
  batch_size = max(feature_lengths.values())
  padded_shapes = {k: [-1] for k in feature_lengths}
  for k in feature_lengths:
    padded_shapes[k].extend(dataset.element_spec[k].shape[1:])
  ds = ds.padded_batch(batch_size, padded_shapes=padded_shapes)

  if use_custom_ops:
    ds = _pack_with_custom_ops(ds, feature_lengths)
  else:
    ds = _pack_with_tf_ops(ds, feature_lengths)

  # Set the Tensor shapes correctly since they get lost in the process.
  def _set_shape(x):
    for k, v in x.items():
      new_shape = [feature_lengths[_strip_packed_feature_key(k)]]
      new_shape.extend(v.get_shape()[1:])
      v.set_shape(new_shape)
    return x

  return ds.map(_set_shape, num_parallel_calls=tf.data.experimental.AUTOTUNE)


def _pack_with_tf_ops(
    dataset: tf.data.Dataset, feature_lengths: Mapping[str, int]
) -> tf.data.Dataset:
  """Helper-function for packing a dataset which has already been batched.

  See trim_and_pack_dataset()

  Uses tf.while_loop. Slow.

  Args:
    dataset: a dataset containing padded batches of examples.
    feature_lengths: mapping from feature key to packed length.

  Returns:
    a dataset.
  """
  empty_example = {}
  for k in feature_lengths:
    for suff in ("", "_positions"):
      empty_example[k + suff] = tf.zeros([0], dtype=tf.int32)
      empty_example[k + suff].set_shape([None])
  keys_etc = empty_example.keys()

  def _write_packed_example(partial, outputs):
    new_partial = empty_example.copy()
    new_outputs = {}
    for k in keys_etc:
      new_outputs[k] = outputs[k].write(
          outputs[k].size(),
          tf.pad(
              partial[k],
              [[
                  0,
                  feature_lengths[_strip_packed_feature_key(k)]
                  - tf.size(partial[k]),
              ]],
          ),
      )
    return new_partial, new_outputs

  def pack_batch(x: Mapping[str, tf.Tensor]) -> Mapping[str, tf.Tensor]:
    """Internal function to map over.

    Consumes a batch of input examples and produces a variable number of output
    examples.

    Args:
      x: a single example

    Returns:
      a tf.data.Dataset
    """
    keys = list(feature_lengths)
    partial = empty_example.copy()
    first_key, *_ = keys
    dynamic_batch_size = tf.shape(x[first_key])[0]
    outputs = {}
    for k in keys:
      outputs[k] = tf.TensorArray(
          tf.int32,
          size=0,
          dynamic_size=True,
          element_shape=[feature_lengths[k]],
      )
      outputs[k + "_positions"] = tf.TensorArray(
          tf.int32,
          size=0,
          dynamic_size=True,
          element_shape=[feature_lengths[k]],
      )

    for i in tf.range(0, dynamic_batch_size):
      tf.autograph.experimental.set_loop_options(
          shape_invariants=[
              (partial, {k: tf.TensorShape([None]) for k in keys_etc}),
              (outputs, {k: tf.TensorShape(None) for k in keys_etc}),
          ]
      )

      can_append = True
      one_example = {}
      for k in keys:
        val = tf.cast(x[k][i], tf.int32)
        val = val[: tf.reduce_sum(tf.cast(tf.not_equal(val, 0), tf.int32))]
        one_example[k] = val
      for k in keys:
        can_append = tf.logical_and(
            can_append,
            tf.less_equal(
                tf.size(partial[k]) + tf.size(one_example[k]),
                feature_lengths[k],
            ),
        )

      if not can_append:
        partial, outputs = _write_packed_example(partial, outputs)

      new_partial = {}
      for k in keys:
        new_seq = one_example[k][: feature_lengths[k]]
        new_seq_len = tf.size(new_seq)
        new_partial[k] = tf.concat([partial[k], new_seq], 0)
        new_partial[k + "_positions"] = tf.concat(
            [partial[k + "_positions"], tf.range(new_seq_len, dtype=tf.int32)],
            0,
        )
      partial = new_partial

    _, outputs = _write_packed_example(partial, outputs)
    packed = {k: outputs[k].stack() for k in keys_etc}
    for k in keys:
      packed[k + "_segment_ids"] = tf.cumsum(
          tf.cast(tf.equal(packed[k + "_positions"], 0), tf.int32), axis=1
      ) * tf.cast(tf.not_equal(packed[k], 0), tf.int32)
    return packed

  dataset = dataset.map(
      pack_batch, num_parallel_calls=tf.data.experimental.AUTOTUNE
  )
  return dataset.unbatch()


def _pack_with_custom_ops(
    dataset: tf.data.Dataset, feature_lengths: Mapping[str, int]
) -> tf.data.Dataset:
  """Helper-function for packing a dataset which has already been batched.

  See trim_and_pack_dataset()

  Relies on custom ops which require a custom compiled binary.
  Faster than _pack_with_tf_ops(), and denser packing.

  Args:
    dataset: a dataset containing padded batches of examples.
    feature_lengths: mapping from feature key to packed length.

  Returns:
    a dataset.
  """
  # TODO(adarob): Move ops into this library and fix int64 issue.
  from tensor2tensor.data_generators.ops import pack_sequences_ops  # pylint: disable=g-import-not-at-top

  keys = list(feature_lengths)
  use_generic_custom_ops = False
  if len(keys) == 1:
    (k1,) = keys
    k2 = k1
  elif len(keys) == 2:
    k1, k2 = keys
  else:
    use_generic_custom_ops = True
    logging.info(
        "`pack_sequences_2` cannot pack more than 2 features. "
        "Using `pack_sequences_k` instead."
    )

  element_spec = dataset.element_spec
  for k in feature_lengths:
    if not element_spec[k].dtype.is_integer:
      use_generic_custom_ops = True
      logging.info(
          (
              "`pack_sequences_2` cannot pack non-integer feature '%s'. "
              "Using `pack_sequences_k` instead."
          ),
          k,
      )
    if not element_spec[k].shape.is_compatible_with(
        tf.TensorShape([None, None])
    ):
      use_generic_custom_ops = True
      logging.info(
          (
              "`pack_sequences_2` cannot pack higher rank feature '%s'. "
              "Using `pack_sequences_k` instead."
          ),
          k,
      )

  def custom_pack_batch(x):
    """Map-function."""
    if use_generic_custom_ops:
      xs = []
      max_lengths = []
      for k in sorted(feature_lengths.keys()):
        xs.append(x[k])
        max_lengths.append(feature_lengths[k])
      (packed, segment_ids, positions) = pack_sequences_ops.pack_sequences_k(
          inputs=xs, max_lengths=max_lengths
      )
      y = {}
      for i, k in enumerate(sorted(feature_lengths.keys())):
        y[k] = packed[i]
        y[f"{k}_segment_ids"] = segment_ids[i]
        y[f"{k}_positions"] = positions[i]
      return y
    logging.info(
        "Features are compatible with `pack_sequences_2`. "
        "Not using `pack_sequences_k`."
    )
    (
        k1_packed,
        k1_segment_ids,
        k1_positions,
        k2_packed,
        k2_segment_ids,
        k2_positions,
    ) = pack_sequences_ops.pack_sequences2(
        # cast to int64 for compatibility with custom ops
        tf.cast(x[k1], tf.int64),
        tf.cast(x[k2], tf.int64),
        feature_lengths[k1],
        feature_lengths[k2],
    )
    packed = {
        k1: k1_packed,
        k1 + "_segment_ids": k1_segment_ids,
        k1 + "_positions": k1_positions,
    }
    if len(keys) == 2:
      packed.update({
          k2: k2_packed,
          k2 + "_segment_ids": k2_segment_ids,
          k2 + "_positions": k2_positions,
      })

    # cast back to int32
    for k, v in packed.items():
      packed[k] = tf.cast(v, tf.int32)

    return packed

  dataset = dataset.map(
      custom_pack_batch, num_parallel_calls=tf.data.experimental.AUTOTUNE
  )
  dataset = dataset.unbatch()
  return dataset


def _shift_right_by_one(tensor: tf.Tensor, bos_id: int = 0) -> tf.Tensor:
  """Shift the input tensor to the right by one position without wrapping."""

  if not (tensor.dtype.is_integer or tensor.dtype.is_floating):
    raise ValueError(f"Only numeric types are supported. Got: {tensor.dtype}")
  # tf.roll wraps around the axis.
  rolled = tf.roll(tensor, shift=1, axis=0)

  # Zero out the first position by multiplying with [0, 1, 1, ..., 1].
  depth = tf.shape(tensor)[0]
  mask = tf.one_hot(0, depth=depth, on_value=0, off_value=1, dtype=tensor.dtype)

  # Expand dims of mask to broadcast to rolled.
  dim_expansion = [slice(None, None)] + [None] * (len(rolled.shape) - 1)
  mask = mask[dim_expansion]
  return rolled * mask + (1 - mask) * bos_id


def make_autoregressive_inputs(
    targets: tf.Tensor,
    sequence_id: tf.Tensor = None,
    output_dtype: Optional[tf.dtypes.DType] = None,
    bos_id: int = 0,
) -> tf.Tensor:
  """Generate inputs for an autoregressive model, by shifting the targets.

  Modified from mesh_tensorflow.transformer.transformer.autoregressive_inputs.

  For the first element of each sequence, the returned input id is 0.

  For a "packed" dataset, also pass the sequence_id tensor, which aligns
  with the targets tensor and contains different values for different
  concatenated examples.

  Example for a packed dataset:

  ```
        targets = [3, 8, 1, 9, 1, 5, 4, 1, 0, 0]
    sequence_id = [1, 1, 1, 2, 2, 3, 3, 3, 0, 0]
         inputs = [0, 3, 8, 0, 9, 0, 5, 4, 0, 0]
                            |     |        |
                            These positions are set to 0 if sequence_id is not
                            None.
  ```

  Args:
    targets: a tf.int32 tensor with shape [length].
    sequence_id: an optional tensor with the same shape as targets.
    output_dtype: an optional output data type.
    bos_id: bos id.

  Returns:
    a tensor with dtype tf.int32 and the same shape as targets.
  """
  output_dtype = output_dtype or targets.dtype
  if sequence_id is not None and not sequence_id.dtype.is_integer:
    raise ValueError(
        "The sequence_id should be integer-valued tensors for a packed dataset."
    )
  if sequence_id is not None and len(targets.shape) > 1:
    raise ValueError(
        "Only 1-D sequences are supported with packing. Got a "
        f"packed {len(targets.shape)}-D sequence."
    )

  inputs = _shift_right_by_one(targets, bos_id)
  if inputs.dtype != output_dtype:
    inputs = tf.cast(inputs, output_dtype)

  # We should have a 0 at the beginning of each sequence rather than the
  # shifted EOS (e.g. 1) from the previous sequence.
  if sequence_id is not None:
    not_first_in_sequence = tf.equal(
        sequence_id, _shift_right_by_one(sequence_id)
    )
    not_first_in_sequence = tf.cast(not_first_in_sequence, output_dtype)
    first_ids = tf.cast((1 - not_first_in_sequence) * bos_id, output_dtype)
    inputs = inputs * not_first_in_sequence + first_ids
  return inputs


# ========================= Mixing Rate Functions ==============================


def mixing_rate_num_examples(
    task,
    maximum: Optional[int] = None,
    scale: float = 1.0,
    temperature: float = 1.0,
    fallback_to_num_input_examples: bool = True,
    split: str = "train",
) -> float:
  """Mixing rate based on the number of examples for the task's split.

  It should be noted that SeqIO only injects the task, and all other parameters
  must be provided by the user when initializing the Mixture.

  Args:
    task: the seqio.Task to compute a rate for.
    maximum: an optional maximum value to clip at after constant scaling but
      before temperature scaling.
    scale: a multiplicative scaling factor applied before temperature.
    temperature: a temperature (T) to scale rate (r) by as r^(1/T).
    fallback_to_num_input_examples: whether to fallback to using the number of
      input examples when the Task is not cached. Otherwise, an error will be
      raised.
    split: the split to look at for cached stats.

  Returns:
    The mixing rate for this task.
  """

  if task.cache_dir or not fallback_to_num_input_examples:
    ret = task.get_cached_stats(split)["examples"]
  else:
    logging.warning(
        (
            "Task '%s' not cached so using number of input examples instead of "
            "preprocessed examples to compute rate."
        ),
        task.name,
    )
    ret = task.num_input_examples(split)

  ret *= scale
  if maximum:
    ret = min(ret, maximum)
  if temperature != 1.0:
    ret = ret ** (1.0 / temperature)
  return ret


def mixing_rate_num_characters(
    task, temperature: float = 1.0, char_count_name: str = "text_chars"
) -> float:
  """Mixing rate based on the number of characters for the task's 'train' split.

  Args:
    task: the seqio.Task to compute a rate for.
    temperature: a temperature (T) to scale rate (r) by as r^(1/T).
    char_count_name: feature name of the character counts in the cached stats
      file.

  Returns:
    The mixing rate for this task.
  """
  if task.cache_dir is None:
    raise ValueError(
        "`mixing_rate_num_characters` requires that each task has is cached "
        "with the character count stats."
    )
  ret = task.get_cached_stats("train")[char_count_name]
  if temperature != 1.0:
    ret = ret ** (1.0 / temperature)
  return ret


# ======================== Decorators =========================================

_NEXT_MAP_SEED = None


@contextlib.contextmanager
def map_seed_manager(initial_seed=None):
  """Contextmanager to control the initial seed used by `map_over_dataset`."""
  global _NEXT_MAP_SEED
  old_map_seed = _NEXT_MAP_SEED
  _NEXT_MAP_SEED = initial_seed
  yield
  _NEXT_MAP_SEED = old_map_seed


def set_preprocessor_seed(preprocessor_fn, seed=None):
  """Sets the internal map seed for the provided preprocessor."""
  return add_kwargs_to_transform(preprocessor_fn, seqio_internal_map_seed=seed)


_SPECIAL_KWARGS = ("sequence_length", "output_features")


@dataclasses.dataclass
class _GrainRandomMapFn(_RandomMapTransform):
  """Grain Transform to represent existing SeqIO random map preprocessors."""

  map_fn: Callable[..., Any]
  num_seeds: int
  num_parallel_calls: int = tf.data.AUTOTUNE

  # These are set by SeqIO before applying the preprocessor or before passing
  # it to Grain.
  sequence_length: Optional[Mapping[str, int]] = None
  output_features: Optional[Mapping[str, Any]] = None

  def _map_fn_with_special_kwargs(self, *args, **kwargs):
    """Handle _SPECIAL_KWARGS before calling self.map_fn()."""
    special_kwargs = {
        k: getattr(self, k)
        for k in _SPECIAL_KWARGS
        if getattr(self, k) is not None
    }
    map_fn = add_kwargs_to_transform(self.map_fn, **special_kwargs)
    return map_fn(*args, **kwargs)

  # Path for Grain. Uses seed provided by Grain.
  def random_map(self, element, rng: tf.Tensor):
    if self.num_seeds == 1:
      return self._map_fn_with_special_kwargs(element, seed=rng)
    rngs = tf.random.experimental.stateless_split(rng, self.num_seeds)
    rngs = tf.unstack(rngs)
    return self._map_fn_with_special_kwargs(element, seeds=rngs)

  # Path for SeqIO; preserves legacy logic to manage seeds and differs in
  # seed-management behavior in Grain.
  def __call__(self, dataset: tf.data.Dataset, *args, **kwargs):
    global _NEXT_MAP_SEED
    if _NEXT_MAP_SEED is None:
      random_ds_seeds = ((None, None),) * self.num_seeds
    else:
      random_ds_seeds = np.arange(
          _NEXT_MAP_SEED, _NEXT_MAP_SEED + 2 * self.num_seeds
      ).reshape(-1, 2)
      random_ds_seeds = tuple(tuple(s) for s in random_ds_seeds)
      _NEXT_MAP_SEED += 2 * self.num_seeds
    seed_datasets = tf.nest.map_structure(
        tf.data.experimental.RandomDataset, random_ds_seeds
    )

    def map_fn(element, seeds):
      if self.num_seeds == 1:
        return self._map_fn_with_special_kwargs(
            element, seed=seeds[0], *args, **kwargs
        )
      return self._map_fn_with_special_kwargs(
          element, seeds=seeds, *args, **kwargs
      )

    return tf.data.Dataset.zip((dataset, seed_datasets)).map(
        map_fn, num_parallel_calls=self.num_parallel_calls
    )


@dataclasses.dataclass
class _GrainMapFn(_MapTransform):
  """Grain Transform to represent existing SeqIO map preprocessors."""

  map_fn: Callable[..., Any]
  num_parallel_calls: int

  # These are set by SeqIO before applying the preprocessor or before passing
  # it to Grain.
  sequence_length: Optional[Mapping[str, int]] = None
  output_features: Optional[Mapping[str, Any]] = None

  def _map_fn_with_special_kwargs(self, *args, **kwargs):
    """Handle _SPECIAL_KWARGS before calling self.map_fn()."""
    special_kwargs = {
        k: getattr(self, k)
        for k in _SPECIAL_KWARGS
        if getattr(self, k) is not None
    }
    map_fn = add_kwargs_to_transform(self.map_fn, **special_kwargs)
    return map_fn(*args, **kwargs)

  # Path for Grain
  def map(self, element):
    return self._map_fn_with_special_kwargs(element)

  # Path for SeqIO
  def __call__(
      self, dataset: tf.data.Dataset, *args, **kwargs
  ) -> tf.data.Dataset:
    return dataset.map(
        lambda x: self._map_fn_with_special_kwargs(x, *args, **kwargs),
        num_parallel_calls=self.num_parallel_calls,
    )


def map_over_dataset(
    fn=None, *, num_seeds=None, num_parallel_calls=tf.data.experimental.AUTOTUNE
):
  """Decorator to map decorated function over dataset.

  Many preprocessors map a function over a dataset. This decorator helps reduce
  boilerplate for this common pattern.

  If `num_seeds` is set to 1, a unique random seed (pair of int32) will be
  passed to the mapping function with keyword 'seed'.
  If `num_seeds` is greater than 1, unique random seeds (pairs of int32) will be
  passed to the mapping function with keyword 'seeds'.
  These seeds can be generated deterministically by using the `map_seed_manager`
  to set the seed for the process that generates the individual seeds for each
  mapping function. These seeds will be set sequentially from the initial seed
  for each call to `map_over_dataset` where `num_seeds > 0`.

  Args:
    fn: map function
    num_seeds: optional number of random seeds (pairs of int32) to pass to the
      mapping function.
    num_parallel_calls: num_parallel_calls value to pass to Dataset.map

  Returns:
    Callable transform which takes dataset as first argument.
  """

  def map_without_seeds(fn):
    @functools.wraps(fn)
    def wrapped_fn(ds, *args, **kwargs):
      return _GrainMapFn(fn, num_parallel_calls)(ds, *args, **kwargs)

    return wrapped_fn

  def map_with_seeds(fn):
    @functools.wraps(fn)
    def wrapped_fn(ds, *args, **kwargs):
      return _GrainRandomMapFn(fn, num_seeds, num_parallel_calls)(
          ds, *args, **kwargs
      )

    return wrapped_fn

  wrapper = map_without_seeds if num_seeds is None else map_with_seeds
  return wrapper if fn is None else wrapper(fn)


def fully_qualified_class_name(instance: Any) -> str:
  """Returns the fully qualified class name of the given instance."""
  return f"{type(instance).__module__}.{type(instance).__name__}"


def function_name(function) -> str:
  """Returns the name of a (possibly partially applied) function."""
  try:
    if inspect.isclass(function):
      # function can be a protocol.
      if hasattr(function.__class__, "__name__"):
        return function.__class__.__name__
      else:
        return str(function.__class__)
    elif isinstance(function, functools.partial):
      # functools.partial can be applied multiple times.
      return function_name(function.func)
    elif hasattr(function, "__name__"):
      return function.__name__
  except Exception:  # pylint: disable=broad-exception-caught
    pass
  # Function name could not be determined.
  return ""





# Copyright 2024 The SeqIO Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Python script to print where TaskRegistry.add calls are coming from.

Prints a TSV to stdout of each registered task (after importing modules from
--module_import), along with the callsites upstream. This is useful for
determining where any TaskRegistry.add calls exist at the module-level (and
are thus run when just importing that module).

Usage:
====================
registration_provenance \
  --registration_module_import=my.tasks
  --registration_provenance_out_filename=/some/file/path.tsv
"""

import importlib
from typing import List

from absl import app
from absl import flags
from seqio import task_registry_provenance_tracking
import tensorflow as tf

FLAGS = flags.FLAGS
flags.DEFINE_multi_string(
    "registration_module_import",
    [],
    (
        "Modules to import. Use this, for example, to add new `Task`s to the "
        "global `TaskRegistry`."
    ),
)

flags.DEFINE_string(
    "registration_provenance_out_filename",
    "/dev/stdout",  # Default is to print to command line.
    (
        "File name to which to write TSV of provenance. Use default to print to"
        " terminal."
    ),
)


def _import_modules(modules):
  """Function that imports an additional module."""
  for module in modules:
    if module:
      importlib.import_module(module)


def _get_lines_for_tsv() -> List[List[str]]:
  """Gets a TSV of provenance for all registered & tracked Tasks/Mixtures.

  Returns:
    List of TSV lines with header line.
    Columns of TSV lines are:
    provider_name    provider_type    filename_1:lineno    filename_2:lineno...

  Raises:
    ValueError if no provenances are found via
    task_registry_provenance_tracking.view_provenances()
  """
  if not task_registry_provenance_tracking.view_provenances():
    raise ValueError(
        "No task/mixture registrations found. Make sure you have turned on"
        " task_registry_provenance_tracking.turn_on_tracking() before importing"
        " any modules that may be registering tasks."
    )

  # The TSV manipulation could be more easily done with a tool like pandas,
  # but we choose instead to avoid bringing in pandas as a dependency for
  # just this small use.
  tsv_lines = []
  for (
      provider_name,
      (provider_type, provenance),
  ) in task_registry_provenance_tracking.view_provenances().items():
    num_columns_for_call_sites = len(provenance)
    current_line = [provider_name, provider_type] + provenance
    tsv_lines.append(current_line)

  header = ["provider_name", "provider_type"] + [
      f"call_site_{i}" for i in range(num_columns_for_call_sites)
  ]
  tsv_lines.insert(0, header)
  return tsv_lines


def _write_lines(tsv_lines: List[List[str]], out_path: str) -> None:
  with tf.io.gfile.GFile(out_path, "w") as f:
    for line in tsv_lines:
      f.write("\t".join(line) + "\n")


def main(_) -> None:
  task_registry_provenance_tracking.turn_on_tracking()
  _import_modules(FLAGS.registration_module_import)

  tsv_lines = _get_lines_for_tsv()
  _write_lines(tsv_lines, FLAGS.registration_provenance_out_filename)


if __name__ == "__main__":
  app.run(main)
